import logging
from fontFeatures import Rule


def shaper_inputs(self):
    """Returns a list of potential glyphs to determine whether to test if this
    rule applies at a given point."""
    return [self.bases.keys(), self.marks.keys()]

def find_base_backwards(self, buf, ix):
    """Looks backwards in a buffer from index ``ix`` to find the nearest base."""
    start_ix = ix
    ix = ix - 1
    while ix >= 0:
        if buf[ix].glyph in self.bases.keys():
            # Check for unhelpful stuff in between
            my_category = buf[ix].category[0]
            for i in range(ix+1, start_ix):
                if buf[i].category[0] == my_category or buf[i].category[0] == "unknown":
                    # Oops, we skipped over another %s to get here
                    return None
            return ix
        ix = ix - 1
    return None

def would_apply_at_position(self, buf, ix, namedclasses={}):
    """Tests to see if this rule would apply at position ``ix`` of the buffer."""
    from fontFeatures.shaperLib.Rule import _expand_slot

    logging.getLogger("fontFeatures.shaperLib").debug("Testing if rule would apply at position %i" % (ix))
    if namedclasses:
        marks = _expand_slot(self.marks.keys(), namedclasses)
        bases = _expand_slot(self.bases.keys(), namedclasses)
    else:
        marks = self.marks.keys()
        bases = self.bases.keys()

    if self.is_cursive:
        if ix == 0:
            logging.getLogger("fontFeatures.shaperLib").debug(" * No, it has no adjacent glyph")
            return False
        # We will sort it out later
        # logging.getLogger("fontFeatures.shaperLib").debug(" * Yes, %s/%s is a pair" % (buf[ix].glyph, buf[ix-1].glyph))
        return True



    # Mark to base is a bit different, as multiple marks can attach to a base
    # so we search backwards for the preceding base glyph
    # XXX mark to mark
    if buf[ix].glyph not in marks:
        logging.getLogger("fontFeatures.shaperLib").debug(" * No, %s is not in our mark list" % (buf[ix].glyph))
        return False
    base_ix = find_base_backwards(self, buf, ix)
    if base_ix is None:
        logging.getLogger("fontFeatures.shaperLib").debug(" * No, I couldn't find a base glyph")
        return False
    if buf[base_ix].glyph not in bases:
        logging.getLogger("fontFeatures.shaperLib").debug(" * No, %s is not in our base list" % buf[base_ix].glyph)
        return False
    logging.getLogger("fontFeatures.shaperLib").debug(" * Yes, attaching mark %s/%i to %s/%i" % (buf[ix].glyph, ix, buf[base_ix].glyph, base_ix))
    return True

def _do_apply_cursive(self, buf, ix):
    this_record = buf[ix]
    prev_record = buf[ix-1]

    mark = prev_record.glyph
    base = this_record.glyph

    if mark not in self.marks or base not in self.bases:
        return
    exit_x, exit_y = self.marks.get(mark, (0,0))
    entry_x, entry_y = self.bases.get(base, (0,0))

    i = ix-1
    j = ix

    if buf.direction == "LTR":
        buf[i].position.xAdvance = exit_x + (buf[i].position.xPlacement or 0)
        d = entry_x + (buf[j].position.xPlacement or 0)
        buf[j].position.xAdvance = (buf[j].position.xAdvance or 0) - d
        buf[j].position.xPlacement = (buf[j].position.xPlacement or 0) - d
    elif buf.direction == "RTL":
        d = exit_x + (buf[i].position.xPlacement or 0)
        logging.getLogger("fontFeatures.shaperLib").debug("Adjusting advance of %s by %i" % (buf[i].glyph, -d))
        buf[i].position.xAdvance = (buf[i].position.xAdvance or 0) - d
        buf[i].position.xPlacement = (buf[i].position.xPlacement or 0) - d

        buf[j].position.xAdvance = entry_x + (buf[j].position.xPlacement or 0)

    child = i
    parent = j
    x_offset = entry_x - exit_x
    y_offset = entry_y - exit_y
    if not (self.flags & 1):
        parent, child = child, parent
        x_offset = -x_offset
        y_offset = -y_offset
    buf[child].position.yPlacement = y_offset
    this_record.attach_type = "cursive"
    this_record.attach_chain = parent - child


def _do_apply(self, buf, ix, namedclasses={}):
    if self.is_cursive:
        return _do_apply_cursive(self, buf, ix)
    from fontFeatures import ValueRecord
    base_ix = find_base_backwards(self, buf, ix)
    mark = buf[ix].glyph
    base = buf[base_ix].glyph
    xpos = self.bases[base][0] - self.marks[mark][0]
    ypos = self.bases[base][1] - self.marks[mark][1]
    buf[ix].position.xPlacement = xpos
    buf[ix].position.yPlacement = ypos
    buf[ix].attach_type = "mark"
    buf[ix].attach_chain = base_ix - ix



"""Routines for converting Attachment rules to and from XML."""

from lxml import etree


def _toXML(self, root):
    root.set("basename", self.base_name)
    root.set("markname", self.mark_name)
    for n, anchor in self.bases.items():
        base = etree.SubElement(root, "base")
        base.set("name", n)
        base.set("anchorX", str(anchor[0]))
        base.set("anchorY", str(anchor[1]))
    for n, anchor in self.marks.items():
        mark = etree.SubElement(root, "mark")
        mark.set("name", n)
        mark.set("anchorX", str(anchor[0]))
        mark.set("anchorY", str(anchor[1]))
    return root


@classmethod
def fromXML(klass, el):
    """Creates a rule from a lxml Element object."""
    rule = klass(
        base_name = el.get("basename"),
        mark_name = el.get("markname"),
        address=el.get("address"),
        flags=int(el.get("flags") or 0),
    )
    for baseormark in el:
        key = baseormark.get("name")
        value = (int(baseormark.get("anchorX")), int(baseormark.get("anchorY")))
        if baseormark.tag == "base":
            rule.bases[key] = value
        elif baseormark.tag == "mark":
            rule.marks[key] = value
    return rule



from youseedee import ucd_data
from .BaseShaper import BaseShaper
import re
from fontFeatures.shaperLib.Buffer import BufferItem
from fontFeatures.shaperLib.VowelConstraints import preprocess_text_vowel_constraints
from .IndicShaperData import script_config, syllabic_category_map, IndicPositionalCategory2IndicPosition, IndicPosition
import unicodedata

DOTTED_CIRCLE = 0x25CC

class SyllabicShaper(BaseShaper):

    basic_features = ['nukt', 'akhn', 'rphf', 'rkrf', 'pref', 'blwf', 'abvf', 'half', 'pstf', 'vatu', 'cjct']
    after_syllable_features = [ "ccmp", "locl" ]
    other_features = [ "init", "pres", "abvs", "blws", "psts", "haln", "calt", "clig" ]
    repha = "Repha"

    def collect_features(self, shaper):
        shaper.add_pause(self.setup_syllables)
        shaper.add_features(*self.after_syllable_features)
        shaper.add_pause(self.initial_reordering)
        for i in self.basic_features:
            shaper.add_features(i)
            shaper.add_pause()
        shaper.add_pause(self.final_reordering)
        shaper.add_features(*self.other_features)

    def preprocess_text(self):
        preprocess_text_vowel_constraints(self.buffer)

    def reassign_category(self, item):
        pass

    def assign_category(self, item):
        # Base behavior is Indic
        ucd = ucd_data(item.codepoint)
        item.syllabic_category = syllabic_category_map.get(ucd.get("Indic_Syllabic_Category", "Other"),"X")
        item.positional_category = ucd.get("Indic_Positional_Category", "x")
        item.syllabic_position = IndicPositionalCategory2IndicPosition(item.positional_category)
        self.reassign_category(item)

    def assign_categories(self):
        serialized = []
        for ix,item in enumerate(self.buffer.items):
            self.assign_category(item)
            serialized.append("<"+item.syllabic_category+">("+item.positional_category+")="+str(ix))
        return "".join(serialized)

    def setup_syllables(self, shaper):
        syllable_index = 0
        category_string = self.assign_categories()
        self.plan.msg("Set up syllables: "+category_string)
        while len(category_string) > 0:
            state, end, matched_type = None, None, None
            for syllable_type in self.syllable_types:
                m = re.match(self.syllable_machine[syllable_type], category_string)
                if m and len(m[0]):
                    matched_type = syllable_type
                    category_string = category_string[len(m[0]):]
                    indexes = re.findall("=(\\d+)", m[0])
                    start, end = int(indexes[0]), int(indexes[-1])
                    break
            assert(matched_type)
            for i in range(start, end+1):
                self.buffer.items[i].syllable_index = syllable_index
                self.buffer.items[i].syllable = syllable_type
            syllable_index = syllable_index+1
        self.plan.msg("Syllables", self.buffer, ["syllable_index", "syllable"])

    def iterate_syllables(self):
        ix = 0
        while ix < len(self.buffer.items):
            syll_type = self.buffer.items[ix].syllable
            index = self.buffer.items[ix].syllable_index
            start = ix
            while ix < len(self.buffer.items) and self.buffer.items[ix].syllable_index == index:
                ix = ix + 1
                end = ix
            yield index, syll_type, start, end

    def insert_dotted_circles(self, repha):
        for ix,i in enumerate(self.buffer.items):
            if i.syllable == "broken_cluster" and (ix == 0 or i.syllable_index != self.buffer.items[ix-1].syllable_index):
                # Need to insert dotted circle.
                dotted_circle = BufferItem.new_unicode(DOTTED_CIRCLE)
                dotted_circle.syllable_index = i.syllable_index
                dotted_circle.syllable = i.syllable
                self.assign_category(dotted_circle)
                dotted_circle.map_to_glyph(self.buffer.font)
                if self.repha is not None and i.syllabic_category == repha:
                    self.buffer.items.insert(ix+1, dotted_circle)
                else:
                    self.buffer.items.insert(ix, dotted_circle)

    def initial_reordering_pre(self):
        pass

    def initial_reordering(self, shaper):
        self.initial_reordering_pre()
        self.insert_dotted_circles(self.repha)
        for index,syll_type,start,end in self.iterate_syllables():
            reorder = self.initial_reordering_syllable.get(syll_type, None)
            if reorder:
                reorder(self, start,end)
        self.plan.msg("After initial reordering", self.buffer, ["syllable_index", "syllable"])

    def final_reordering(self, shaper):
        for index,syll_type,start,end in self.iterate_syllables():
            self.final_reordering_syllable(start,end)

    def final_reordering_syllable(self, start, end):
        pass




from .IndicShaperData import IndicPosition, make_syllable_machine, syllabic_category_map
from .SyllabicShaper import SyllabicShaper
import unicodedata
from youseedee import ucd_data
from collections import OrderedDict


khmer_category_reassignments = {
    0x179A: "Ra",
    0x17CC: "Robatic",
    0x17C9: "Robatic",
    0x17CA: "Robatic",
    0x17C6: "Xgroup",
    0x17CB: "Xgroup",
    0x17CD: "Xgroup",
    0x17CE: "Xgroup",
    0x17CF: "Xgroup",
    0x17D0: "Xgroup",
    0x17D1: "Xgroup",
    0x17C7: "Ygroup",
    0x17C8: "Ygroup",
    0x17DD: "Ygroup",
    0x17D3: "Ygroup",
}

states = OrderedDict(
    c = "(C | Ra | V)",
    cn = "c ((ZWJ|ZWNJ)? Robatic)?",
    joiner = "(ZWJ | ZWNJ)",
    xgroup = "(joiner* Xgroup)*",
    ygroup = "Ygroup*",
    matra_group = "VPre? xgroup VBlw? xgroup (joiner? VAbv)? xgroup VPst?",
    syllable_tail = "xgroup matra_group xgroup (Coeng c)? ygroup",
    broken_cluster = "(Coeng cn)* (Coeng | syllable_tail)",
    consonant_syllable = "(cn|PLACEHOLDER|DOTTEDCIRCLE) broken_cluster",
)


class KhmerShaper(SyllabicShaper):
    repha = "Repha"
    syllable_machine = make_syllable_machine(
        states,
        additional_categories=["Ra","Robatic","Xgroup", "Ygroup", "VPre", "VBlw", "VAbv", "VPst"]
    )
    syllable_types = [ "consonant_syllable", "broken_cluster", "other" ]

    def collect_features(self, shaper):
        shaper.add_pause(self.setup_syllables)
        shaper.add_pause(self.initial_reordering)
        shaper.add_features("locl", "ccmp")
        shaper.add_features("pref", "blwf", "abvf", "pstf", "cfar")
        shaper.add_pause() # Clear syllables
        shaper.add_features("pres", "blws", "abvs", "psts")


    def override_features(self, shaper):
        shaper.add_features("clig")
        shaper.disable_feature("liga")

    def reassign_category(self, item):
        cp = item.codepoint
        if cp in khmer_category_reassignments:
            item.syllabic_category = khmer_category_reassignments[cp]

        if item.syllabic_category == "M":
            if item.syllabic_position == IndicPosition.PRE_C:
                item.syllabic_category = "VPre"
            elif item.syllabic_position == IndicPosition.ABOVE_C:
                item.syllabic_category = "VAbv"
            elif item.syllabic_position == IndicPosition.BELOW_C:
                item.syllabic_category = "VBlw"
            elif item.syllabic_position == IndicPosition.POST_C:
                item.syllabic_category = "VPst"

    def initial_reordering_consonant_syllable(self, start, end):
        def cat(i):
            return self.buffer.items[i].syllabic_category
        def mask_disallow(places, *feats):
            for p in places:
                for f in feats:
                    self.buffer.items[p].feature_masks[f] = True
        def mask_allow(places, *feats):
            for p in places:
                for f in feats:
                    self.buffer.items[p].feature_masks[f] = False

        mask_allow(range(start+1, end), "blwf", "advf", "pstf")
        mask_disallow(range(start, end), "cfar")

        num_coengs = 0
        for i in range(start+1, end):
            if cat(i) == "Coeng" and num_coengs <= 2 and i + 1 < end:
                num_coengs += 1
                if cat(i+1) == "Ra":
                    mask_allow([i, i+1], "pref")
                    # Move the Coeng,Ro sequence to the start.
                    # XXX merge_clusters
                    ro = self.buffer.items.pop(i+1)
                    coeng = self.buffer.items.pop(i)
                    self.buffer.items = self.buffer.items[:start] + [coeng, ro] + self.buffer.items[start:]
                    mask_allow(range(i+2, end), "cfar")
                    num_coengs = 2
            elif cat(i) == "VPre":
                # XXX merge_clusters
                vpre = self.buffer.items.pop(i)
                self.buffer.items = self.buffer.items[:start] + [vpre] + self.buffer.items[start:]

    initial_reordering_syllable = {
        "broken_cluster": initial_reordering_consonant_syllable,
        "consonant_syllable": initial_reordering_consonant_syllable,
    }


    def normalize_unicode_buffer(self):
        unicodes = [item.codepoint for item in self.buffer.items]
        newunicodes = []
        for cp in unicodes:
            if cp in [0x17BE,0x17BF, 0x17C0, 0x17C4, 0x17C5]:
                newunicodes.extend([0x17C1, cp])
            else:
                newunicodes.extend([ord(x) for x in unicodedata.normalize("NFD", chr(cp)) ])
        # Now recompose
        newstring = ""
        ix = 0
        while ix < len(newunicodes):
            a = newunicodes[ix]
            if ix+1 == len(newunicodes):
                newstring = newstring + chr(a)
                break

            b = newunicodes[ix+1]
            s = chr(a) + chr(b)
            composed = unicodedata.normalize("NFC", s)
            if ucd_data(a)["General_Category"][0] == "M":
                newstring = newstring + chr(a)
                ix = ix + 1
                continue
            elif composed != unicodedata.normalize("NFD", s):
                assert(len(s) == 1)
                newunicodes[ix] = ord(x)
                del newunicodes[ix+1]
                continue
            else:
                newstring = newstring + chr(a)
                ix = ix + 1

        self.buffer.store_unicode(newstring)



import itertools
from fontFeatures.ttLib import unparse
from argparse import ArgumentParser
from fontTools.ttLib import TTFont


parser = ArgumentParser()
parser.add_argument("input", help="font file to process", metavar="FILE")
args = parser.parse_args()
font = TTFont(args.input)
parsed = unparse(font)

flatten = itertools.chain.from_iterable
mark = list(flatten([x.rules for x in parsed.features["mark"]]))
mkmk = list(flatten([x.rules for x in parsed.features["mkmk"]]))
allfeat = mark + mkmk

import code; code.interact(local=locals())

for f in allfeat:
    for b in f.bases.keys():
        if not b in parsed.anchors:
            parsed.anchors[b] = {}
        parsed.anchors[b][f.base_name] = f.bases[b]
    for b in f.marks.keys():
        if not b in parsed.anchors:
            parsed.anchors[b] = {}
        parsed.anchors[b][f.mark_name] = f.marks[b]

for glyph in parsed.anchors.keys():
    print("Anchors %s {" % glyph)
    for name, anchor in parsed.anchors[glyph].items():
        print("\t%s <%i %i>" % (name, *anchor))
    print("};")



"""Convert fontFeatures.Routine objects to binary GPOS/GSUB lookups."""
import fontTools.otlLib.builder as otl
from fontTools.varLib.builder import buildVarDevTable
from fontTools.feaLib.variableScalar import VariableScalar
import itertools


def toOTLookup(self, font, ff):
    """Converts a fontFeatures.Routine object to binary.

    Args:
        font: A ``TTFont`` object.
        ff: The parent ``FontFeatures`` object containing this routine.

    Returns a list of ``fontTools.otlLib.builder`` Builder objects allowing this
    routine to be converted to binary layout format.
    """

    lookuptypes = [x.lookup_type() for x in self.rules]
    if not all([lu == lookuptypes[0] for lu in lookuptypes]):
        raise ValueError("For now, a routine can only contain rules of the same type")

    if not all([self.rules[0].flags == rule.flags for rule in self.rules]):
        raise ValueError("For now, a routine can only contain rules of the same flags")
    self.flags = self.rules[0].flags

    if self.stage == "pos":
        return buildPos(self, font, lookuptypes[0], ff)
    if self.stage == "sub":
        return buildSub(self, font, lookuptypes[0], ff)


def makeAnchor(anchor, ff):
    """Creates an OpenType anchor object from. If the anchor position contains
     a variable scalar, these variable scalars are saved in the ``GDEF`` variation
     store.

     Args:
        anchor: a tuple of x,y position. These elements may be numbers or
            :py:class:`VariableScalar` objects.

    Returns: an ``otTables.Anchor`` object.
    """
    x, y = anchor
    if isinstance(x, VariableScalar):
        x_def, x_index = x.add_to_variation_store(ff.varstorebuilder)
    else:
        x_def, x_index = int(x), None
    if isinstance(y, VariableScalar):
        y_def, y_index = y.add_to_variation_store(ff.varstorebuilder)
    else:
        y_def, y_index = int(y), None
    anchor = otl.buildAnchor(x_def, y_def)
    if x_index is not None and x_index != 0xFFFFFFFF:
        anchor.XDeviceTable = buildVarDevTable(x_index)
        anchor.Format = 3
    if y_index is not None and x_index != 0xFFFFFFFF:
        anchor.YDeviceTable = buildVarDevTable(y_index)
        anchor.Format = 3
    return anchor


def buildPos(self, font, lookuptype, ff):
    """Build a GPOS subtable."""
    builders = []
    if lookuptype == 1:
        builder = otl.SinglePosBuilder(font, self.address)
        for rule in self.rules:
            ot_valuerecs = [
                x.toOTValueRecord(ff, pairPosContext=False) for x in rule.valuerecords
            ]
            for glyph in rule.glyphs[0]:
                builder.add_pos(rule.address, glyph, ot_valuerecs[0])
    elif lookuptype == 2:
        builder = otl.PairPosBuilder(font, self.address)
        for rule in self.rules:
            ot_valuerecs = [
                x.toOTValueRecord(ff, pairPosContext=True) for x in rule.valuerecords
            ]
            if len(rule.glyphs[0]) == 1 and len(rule.glyphs[1]) == 1:
                builder.addGlyphPair(
                    rule.address,
                    rule.glyphs[0][0],
                    ot_valuerecs[0],
                    rule.glyphs[1][0],
                    ot_valuerecs[1],
                )
            else:
                builder.addClassPair(
                    rule.address,
                    tuple(rule.glyphs[0]),
                    ot_valuerecs[0],
                    tuple(rule.glyphs[1]),
                    ot_valuerecs[1],
                )
    elif lookuptype == 3:
        builder = otl.CursivePosBuilder(font, self.address)
        for rule in self.rules:
            allglyphs = set(rule.bases.keys()) | set(rule.marks.keys())
            for g in allglyphs:
                builder.attachments[g] = (
                    g in rule.bases and makeAnchor(rule.bases[g], ff) or None,
                    g in rule.marks and makeAnchor(rule.marks[g], ff) or None,
                )
    elif lookuptype == 4 or lookuptype == 6:
        if lookuptype == 4:
            builder = otl.MarkBasePosBuilder(font, self.address)
            baseholder = builder.bases
        else:
            builder = otl.MarkMarkPosBuilder(font, self.address)
            baseholder = builder.baseMarks
        for r in self.rules:
            for mark, anchor in r.marks.items():
                if mark in builder.marks:
                    raise ValueError(
                        "A mark glyph %s tried to be in two categories (%s, %s)"
                        % (mark, r.mark_name, builder.marks[mark][0])
                    )
                builder.marks[mark] = (r.mark_name, makeAnchor(anchor, ff))
            for base, anchor in r.bases.items():
                if base not in baseholder:
                    baseholder[base] = {}
                baseholder[base][r.mark_name] = makeAnchor(anchor, ff)
        # XXX. We may need to express this as multiple lookups
    elif lookuptype == 8:
        builder = otl.ChainContextPosBuilder(font, self.address)
        for r in self.rules:
            new_lookup_list = []
            import fontFeatures
            if isinstance(r, fontFeatures.Positioning):
                lookups = []
                for glyphs, vr in zip(r.glyphs, r.valuerecords):
                    # Make a fake pos routine
                    subbuilder = buildPos(fontFeatures.Routine(rules=[fontFeatures.Positioning(
                        [glyphs],
                        valuerecords = [vr]
                    )]), font, 1, ff)
                    builders.extend(subbuilder)
                    new_lookup_list.append(subbuilder)
                glyphs = r.glyphs
            else:
                for list_of_lookups in r.lookups:
                    new_lookup_list.append(
                        [lu.routine.__builder for lu in (list_of_lookups or [])]
                    )
                glyphs = r.input
            builder.rules.append(
                otl.ChainContextualRule(
                    r.precontext or [],
                    glyphs or [],
                    r.postcontext or [],
                    new_lookup_list,
                )
            )
    else:
        raise ValueError("Don't know how to build a POS type %i lookup" % lookuptype)
    builder.lookupflag = self.flags
    # XXX mark filtering set
    self.__builder = builder
    builders.append(builder)
    return builders


def buildSub(self, font, lookuptype, ff):
    """Build a GSUB subtable."""
    builders = []
    if lookuptype == 1:
        builder = otl.SingleSubstBuilder(font, self.address)
        for rule in self.rules:
            for left, right in zip(rule.input[0], rule.replacement[0]):
                builder.mapping[left] = right
    elif lookuptype == 2:
        builder = otl.MultipleSubstBuilder(font, self.address)
        for rule in self.rules:
            builder.mapping[rule.input[0][0]] = [x[0] for x in rule.replacement]
    elif lookuptype == 4:
        builder = otl.LigatureSubstBuilder(font, self.address)
        for rule in self.rules:
            for sequence in itertools.product(*rule.input):
                builder.ligatures[sequence] = rule.replacement[0][0]
    elif lookuptype == 6:
        builder = otl.ChainContextSubstBuilder(font, self.address)
        for r in self.rules:
            import fontFeatures
            if isinstance(r, fontFeatures.Substitution) and r.replacement:
                lookups = []
                for left, right in zip(r.input, r.replacement):
                    # Make a fake sub routine

                    subbuilder = buildSub(fontFeatures.Routine(rules=[fontFeatures.Substitution(
                        [left],
                        replacement = [right]
                    )]), font, 1, ff)
                    builders.extend(subbuilder)
                    lookups.append(subbuilder)
            else:
                lookups = []
                for list_of_lookups in r.lookups:
                    lookups.append(
                        [lu.routine.__builder for lu in (list_of_lookups or [])]
                    )
            builder.rules.append(
                otl.ChainContextualRule(
                    r.precontext or [],
                    r.input or [],
                    r.postcontext or [],
                    lookups,
                )
            )
    elif lookuptype == 8:
        builder = otl.ReverseChainSingleSubstBuilder(font, self.address)
        for rule in self.rules:
            mapping = { l[0]:r[0] for l,r in zip(rule.input, rule.replacement) }
            builder.rules.append((rule.precontext, rule.postcontext, mapping))
    else:
        raise ValueError("Don't know how to build a SUB type %i lookup" % lookuptype)

    builder.lookupflag = self.flags
    # XXX mark filtering set
    self.__builder = builder
    builders.append(builder)
    return builders



def shaper_inputs(self):
    return self.glyphs


def _do_apply(self, buf, ix, namedclasses={}):
    coverage = buf[ix : ix + len(self.glyphs)]
    for glyph, vr in zip(coverage, self.valuerecords):
    	glyph.add_position(vr)
    if not self.valuerecords[-1]: # A special case!
        return
    return len(self.glyphs) - 1



"""ttLib.Chaining: Converting Chaining rules to TrueType."""

def lookup_type(self):
    """Mixin to determine the GSUB/GPOS lookup type of a fontFeatures.Chaining object

    Returns: integer GSUB/GPOS lookup type."""
    if self.stage == "pos":
        return 8
    else:
        return 6



from fontFeatures import Substitution, FontFeatures
from fontTools.ttLib import TTFont
from fontFeatures.ttLib.GPOSUnparser import GPOSUnparser
from fontFeatures.ttLib import unparseLanguageSystems
import pprint
import unittest


class TestUnparse(unittest.TestCase):
    font = TTFont("fonts/Amiri-Regular.ttf")
    lookups = font["GPOS"].table.LookupList.Lookup
    unparser = GPOSUnparser(font["GPOS"], None, [])
    unparser.font = font

    def test_single_f1(self):
        g, _ = self.unparser.unparseLookup(self.lookups[1], 1)  # part of kern
        self.assertEqual(g.rules[0].asFea(), "pos uni0621.float <-146 0 0 0>;")

    def test_single_f1_class(self):
        g, _ = self.unparser.unparseLookup(self.lookups[3], 3)  # part of kern
        self.assertEqual(
            g.rules[0].asFea(),
            "pos [uni0627.fina uni0627.fina_KafMemAlf uni0627.fina_MemAlfFina uni0627.fina_Wide] <122 0 122 0>;",
        )

    def test_single_f2(self):
        g, _ = self.unparser.unparseLookup(
            self.lookups[3], 3
        )  # lookups called in mark feature
        self.assertEqual(g.rules[1].asFea(), "pos uni0627 <73 0 146 0>;")
        self.assertEqual(g.rules[2].asFea(), "pos uni06E5.low <24 0 122 0>;")

    def test_pair_f1(self):
        g, _ = self.unparser.unparseLookup(
            self.lookups[56], 56
        )  # proportional digits in kern feature
        self.assertEqual(g.rules[0].asFea(), "pos uni0660.prop uni0667.prop 24;")
        self.assertEqual(g.rules[1].asFea(), "pos uni0660.prop uni0668.prop 24;")

    def test_pair_f2(self):
        g, _ = self.unparser.unparseLookup(self.lookups[76], 76)  # kerns
        self.assertEqual(
            g.rules[0].asFea(),
            "pos [zero zero.prop] [A Aacute Abreve Acircumflex Adieresis Agrave Amacron Aogonek Aring Atilde] -10;",
        )
        self.assertEqual(
            g.rules[1].asFea(),
            "pos [zero zero.prop] [Y Yacute Ycircumflex Ydieresis Ygrave] -21;",
        )

    # def test_pair(self):
    #   g,_ = self.unparser.unparseLookup(self.lookups[56])
    #   self.assertEqual(g.rules[0].asFea(),"pos uni0621.float <-146 0 0 0>;")



"""GDEF Unparser: Extract fontFeatures information from GDEF table."""
from fontTools.misc.py23 import *
import fontTools
from fontTools.feaLib.ast import *
from collections import OrderedDict


class GDEFUnparser:
    """Extract GDEF information into a fontFeatures object.

    Args:
        table: a GDEF table (e.g. ``font["GDEF"]``)
        ff: a fontFeatures object.
    """
    def __init__(self, table, fontfeatures):
        self.table = table.table
        self.ff = fontfeatures

    def unparse(self):
        """Extracts glyph class definitions and stores them in the fontFeatures
        object.

        In theory other information can be retrieved, but currently there is no
        place to store it in the fontFeatures object.
        """
        # if self.table.MarkAttachClassDef is not None:
        #     self.unparseMarkAttachClassDefinitions()

        # if (
        #     hasattr(self.table, "MarkGlyphSetsDef")
        #     and self.table.MarkGlyphSetsDef is not None
        # ):
        #     self.unparseMarkGlyphSetDefinitions()

        if self.table.GlyphClassDef is not None:
            self.unparseGlyphClassDefinitions()

        # if self.table.AttachList is not None:
        #     self.unparseAttachList()

        # if self.table.LigCaretList is not None:
        #     self.unparseLigCaretList()

    # def unparseMarkAttachClassDefinitions(self):
    #     self.preamble.append(Comment("# Mark attachment classes\n"))
    #     markAttachClassDef = self.table.MarkAttachClassDef
    #     attachClasses = {}

    #     for key, val in markAttachClassDef.classDefs.items():
    #         if not val in attachClasses:
    #             attachClasses[val] = GlyphClass()
    #         attachClasses[val].append(key)

    #     for key, val in attachClasses.items():
    #         self.preamble.append(
    #             GlyphClassDefinition("markClass_" + str(key), val)
    #         )

    # def unparseMarkGlyphSetDefinitions(self):
    #     lines = []
    #     self.preamble.append(Comment("# Mark glyph set classes\n"))
    #     for idx, coverage in enumerate(self.table.MarkGlyphSetsDef.Coverage):
    #         glyphclass = GlyphClass()
    #         for g in coverage.glyphs:
    #             glyphclass.append(g)
    #         self.preamble.append(
    #             GlyphClassDefinition("markGlyphSet_" + str(idx), glyphclass)
    #         )

    def unparseGlyphClassDefinitions(self):
        """Fills the fontFeatures ``glyphclasses`` dictionary from the
        ``GlyphClassDef`` subtable."""
        classDefs = self.table.GlyphClassDef.classDefs
        classNames = {
            1: "base",  # Base glyph (single character, spacing glyph)
            2: "ligature",  # Ligature glyph (multiple character, spacing glyph)
            3: "mark",  # Mark glyph (non-spacing combining glyph)
            4: "component",  # Component glyph (part of single character, spacing glyph)
        }
        for key, val in classDefs.items():
            self.ff.glyphclasses[key] = classNames[val]

    # def unparseAttachList(self):
    #     for idx, glyph in enumerate(self.table.AttachList.Coverage.glyphs):
    #         attachmentPoints = self.table.AttachList.AttachPoint[idx]
    #         self.feature.statements.append(
    #             AttachStatement(GlyphName(glyph), attachmentPoints.PointIndex)
    #         )

    # def unparseLigCaretList(self):
    #     ligCaretList = self.table.LigCaretList
    #     for idx, glyph in enumerate(ligCaretList.Coverage.glyphs):
    #         glyph = GlyphName(glyph)
    #         ligGlyph = ligCaretList.LigGlyph[idx]
    #         formats = list({cv.Format for cv in ligGlyph.CaretValue})
    #         assert (
    #             len(formats) == 1
    #         ), "Can't use different CaretValue formats in one LigatureCaret entry"
    #         caretsFormat = formats[0]
    #         if caretsFormat == 1:
    #             # Format == 1: Design units only
    #             # LigatureCaretByPos <glyph|glyphclass> <caret position value>+;
    #             values = [str(cv.Coordinate) for cv in ligGlyph.CaretValue]
    #             self.feature.statements.append(
    #                 LigatureCaretByPosStatement(glyph, values)
    #             )
    #         elif caretsFormat == 2:
    #             # Format == 2: Contour point
    #             values = [str(cv.CaretValuePoint) for cv in ligGlyph.CaretValue]
    #             self.feature.statements.append(
    #                 LigatureCaretByIndexStatement(glyph, values)
    #             )
    #         elif caretsFormat == 3:
    #             # Format == 3: Design units plus Device table
    #             self.feature.statements.append(
    #                 Comment("# LigatureCaretByDev found, but not implemented")
    #             )



import re
from fontFeatures import *
from fontFeatures.optimizer import Optimizer
from fontTools.ttLib import TTFont
from collections import OrderedDict
from fontTools.feaLib.ast import ValueRecord
import warnings


class FontDameParser:
    """Convert layout files in Monotype's FontDame format to fontFeatures.

    Args:
        lines: An array of strings containing the FontDame file, one line per string.
        config: A dictionary of glyph class names.
        glyphset: A list of glyph names in the font.
    """
    def __init__(self, lines, config={}, glyphset=()):
        self.all_languages = []
        self.lines = lines
        self.script_applications = {}
        self.features = {}
        self.lookups = OrderedDict()
        self.dependencies = {}
        self.classes = []
        self.config = config
        self.glyphset = glyphset
        self.current_lookup = None
        self.state = "doing_nothing"
        self.ff = FontFeatures()
        self.resetContexts()

    def resetContexts(self):
        self.classContexts = {}
        self.backtrackclassContexts = {}
        self.lookaheadclassContexts = {}

    def parse(self):
        """Parses the font file, creating a fontFeatures object.

        Returns: A fontFeatures object containing the rules in the FontDame file."""
        # Parse lookups
        for line in self.lines:
            self.parse_line(line)

        # Tidy up lookups
        for lid, lu in self.lookups.items():
            for rule in lu.rules:
                if not isinstance(rule, Chaining):
                    continue
                pretendlookups = rule.lookups
                reallookups = [None] * len(rule.input)
                for i in pretendlookups:
                    m = re.match("\\s*(\\S+),\\s*(\\S+)", i)
                    if not m:
                        raise ValueError("Unparsable lookup chain %s" % i)
                    if not lid in self.dependencies:
                        self.dependencies[lid] = []
                    self.dependencies[lid].append(m[2])
                    if not reallookups[int(m[1]) - 1]:
                        reallookups[int(m[1]) - 1] = []
                    reallookups[int(m[1]) - 1].append(self.lookups[m[2]])
                rule.lookups = reallookups

        # Tie features to languages
        for k, v in self.features.items():
            v["languages_and_scripts"] = self.script_applications[int(k)]

        # Rearrange into features
        self.base_lu_for_feature = {}
        self.toplevel_lookups = set()
        for i in sorted(self.features.keys()):
            feat = self.features[i]
            tag = feat["tag"]
            if "DFLT/default" in feat["languages_and_scripts"]:
                self.base_lu_for_feature[tag] = set(feat["lookups"])
                lookups = [self.lookups[x] for x in feat["lookups"]]
            else:
                # Set difference
                if tag in self.base_lu_for_feature:
                    feat["lookups"] = [
                        item
                        for item in feat["lookups"]
                        if item not in self.base_lu_for_feature[tag]
                    ]
                lookups = [self.lookups[x] for x in feat["lookups"]]
                langcode = [
                    self.format_langcode(x.split("/"))
                    for x in feat["languages_and_scripts"]
                ]
                # Clone the routines just in case
                lookups = [
                    Routine(languages=langcode, rules=lu.rules, flags=lu.flags)
                    for lu in lookups
                ]

            self.ff.addFeature(tag, lookups)
            for lu in feat["lookups"]:
                self.toplevel_lookups.add(lu)

        # Delete toplevel lookups from lookup table
        for l in self.toplevel_lookups:
            del self.lookups[l]

        # Rearrange lookups into dependency order
        done = {}

        def dolookup(lid):
            if lid in done:
                return
            if lid in self.dependencies:
                for x in self.dependencies[lid]:
                    dolookup(x)
            self.ff.routines.append(self.lookups[lid])
            done[lid] = True

        for i in self.lookups.keys():
            dolookup(i)
        return self.ff

    def format_langcode(self, code):
        script, lang = code
        if lang == "default":
            return (script, "dflt")
        return (script, lang)

    def parse_line(self, line):
        if line == "\n":
            return
        elif line == "script table begin\n":
            self.state = "reading_script_table"
            return
        elif line == "feature table begin\n":
            self.state = "reading_feature_table"
            return
        elif line == "class definition begin\n":
            self.state = "reading_class_definition"
            return
        elif line == "backtrackclass definition begin\n":
            self.state = "reading_backtrackclass_definition"
            return
        elif line == "lookaheadclass definition begin\n":
            self.state = "reading_lookaheadclass_definition"
            return

        elif line == "class definition end\n":
            self.state = "parsing_lookup"
            return
        elif line == "feature table end\n":
            self.state = "doing_nothing"
            return
        elif line.startswith("lookup end"):
            self.end_lookup()
            self.state = "doing_nothing"
            return
        elif line == "script table end\n":
            self.state = "doing_nothing"
            return
        elif line.startswith("lookup"):
            self.parse_lookup_header(line)
            self.state = "parsing_lookup"
            return
        if line.startswith("#") or line.startswith("%"):
            return

        if self.state == "reading_script_table":
            self.add_to_script_table(line)
        elif self.state == "reading_feature_table":
            self.add_to_feature_table(line)
        elif self.state == "parsing_lookup":
            self.add_to_lookup(line)
        elif self.state == "reading_class_definition":
            self.add_to_class_definition("class", line)
        elif self.state == "reading_backtrackclass_definition":
            self.add_to_class_definition("backtrackclass", line)
        elif self.state == "reading_lookaheadclass_definition":
            self.add_to_class_definition("lookaheadclass", line)

    def add_to_script_table(self, line):
        m = re.match("^(\\w+)\\s+(\\w+)\\s+(.*)$", line)
        lang = m[1] + "/" + m[2]
        self.all_languages.append(lang)
        for f in re.split(r",\s*", m[3]):
            f = int(f)
            if not (f in self.script_applications):
                self.script_applications[f] = []
            self.script_applications[f].append(lang)

    def add_to_feature_table(self, line):
        m = re.match("^(\w+)\s+(\w+)\s+(.*)$", line)
        self.features[int(m[1])] = {
            "tag": m[2],
            "lookups": re.split(r",\s*", m[3]),
        }

    def end_lookup(self):
        # print("Parsed lookup %s" % self.current_lookup.name)
        # Optimizer().optimize_routine(self.current_lookup)
        self.current_lookup = None
        self.resetContexts()

    def parse_lookup_header(self, line):
        m = re.match("^lookup\s+([\\w-]+)\s+(.*)$", line)
        if not m:
            raise ValueError("Unparsable lookup header: |%s|" % line)
        self.current_lookup = Routine(name="lookup_%s" % m[1])
        self.lookups[m[1]] = self.current_lookup
        self.current_lookup_type = m[2]

    def get_class(self, cid, lookup):
        res = lookup["classes"][cid]
        if len(res) == 1:
            return res[0]
        if len(res) > 5:
            if not tuple(res) in self.classes:
                self.classes.append(tuple(res))
            classname = "@class%i" % self.classes.index(tuple(res))
            if classname in self.config:
                return self.config[classname]
            return classname

        return res

    def append_lookup_flag(self, flag):
        if flag == "IgnoreMarks":
            parsedflag = 8
        elif flag == "IgnoreLigatures":
            parsedflag = 4
        elif flag == "IgnoreBaseGlyphs":
            parsedflag = 2
        elif flag == "RightToLeft":
            parsedflag = 1
        elif re.match("^mat(\\d+)$", flag):
            parsedflag = int(flag[3:]) << 8
        else:
            print("Unknown flag %s" % flag)
        self.current_lookup.flags = self.current_lookup.flags | parsedflag

    def add_subst(self, in_, out_):
        self.current_lookup.addRule(Substitution(in_, out_))

    def add_chain_simple(self, context, lookups):
        self.current_lookup.addRule(Chaining(context, lookups=lookups))

    def add_cursive(self, entryexit, glyph, pos):
        if len(self.current_lookup.rules) == 0:
            self.current_lookup.addRule(Attachment("cursive_entry", ""))
        rule = self.current_lookup.rules[-1]
        if entryexit == "entry":
            rule.bases[glyph] = pos
        else:
            rule.marks[glyph] = pos

    def add_single_pos_x(self, glyph, xa):
        self.current_lookup.addRule(Positioning([[glyph]], [ValueRecord(xAdvance=xa)]))
    def add_single_pos_x_placement(self, glyph, xp):
        self.current_lookup.addRule(Positioning([[glyph]], [ValueRecord(xPlacement=xp)]))
    def add_single_pos_y_placement(self, glyph, yp):
        self.current_lookup.addRule(Positioning([[glyph]], [ValueRecord(xPlacement=yp)]))

    def add_pair(self, glyph1, glyph2, xa):
        self.current_lookup.addRule(Positioning([[glyph1], [glyph2]], [ValueRecord(xPlacement=xa),ValueRecord()]))

    def add_attach(self, entryexit, glyph, pos):
        if len(self.current_lookup.rules) == 0:
            self.current_lookup.addRule(
                Attachment(
                    self.current_lookup.name + "_BS", self.current_lookup.name + "_MK"
                )
            )
        rule = self.current_lookup.rules[-1]
        if entryexit == "base":
            rule.bases[glyph] = pos
        else:
            rule.marks[glyph] = pos

    def add_to_lookup(self, line):
        m = re.match("(\w+)\s+(yes|no)", line)
        if m:
            if m[2] == "yes":
                self.append_lookup_flag(m[1])
            return
        m = re.match("MarkAttachmentType\s+(\d+)", line, flags=re.IGNORECASE)
        if m:
            self.append_lookup_flag("mat" + m[1])
            return

        if self.current_lookup_type == "single":
            m = re.match(r"x advance\s+([\w\.-]+)\s+(-?\d+)\n", line)
            if m:
                self.add_single_pos_x(m[1], int(m[2]))
                return
            m = re.match(r"y placement\s+([\w\.-]+)\s+(-?\d+)\n", line)
            if m:
                self.add_single_pos_y_placement(m[1], int(m[2]))
                return
            m = re.match(r"x placement\s+([\w\.-]+)\s+(-?\d+)\n", line)
            if m:
                self.add_single_pos_x_placement(m[1], int(m[2]))
                return
            m = re.match("([\\w\\.-]+)\s+([\\w\\.-]+)\n", line)
            if not m:
                warnings.warn("Odd single lookup: %s" % line)
            #     import code

            #     code.interact(local=locals())
            else:
                self.add_subst([[m[1]]], [[m[2]]])
        elif self.current_lookup_type == "pair":
            m = re.match(r"left x advance\s+([\w\.-]+)\s+([\w\.-]+)\s+(-?\d+)\n", line)
            if not m:
                warnings.warn("Odd pair lookup: %s" % line)
            else:
                self.add_pair(m[1], m[2], int(m[3]))

        elif self.current_lookup_type == "multiple":
            m = re.match("([\\w\\.-]+)\s+(.*)\n", line)
            self.add_subst([[m[1]]], [[x] for x in m[2].split("\t")])

        elif self.current_lookup_type == "ligature":
            m = re.match("([\\w\\.-]+)\s(.*)\n", line)
            if not m:
                raise ValueError("Unparsable line '%s'" % line)
            self.add_subst([[m[2]]], [m[1].split("\t")])

        elif self.current_lookup_type == "cursive":
            m = re.match(r"(entry|exit)\s+(\S+)\s+(-?\d+),(-?\d+).*\n", line)
            if not m:
                raise ValueError("Unparsable cursive '%s'" % line)
            self.add_cursive(m[1], m[2], (int(m[3]), int(m[4])))

        elif (
            self.current_lookup_type == "mark to base"
            or self.current_lookup_type == "mark to mark"
        ):
            m = re.match(r"(mark|base)\s+(\S+)\s+\S+\s+(-?\d+),(-?\d+).*\n", line)
            if not m:
                raise ValueError("Unparsable mark '%s'" % line)
            self.add_attach(m[1], m[2], (int(m[3]), int(m[4])))

        elif self.current_lookup_type == "mark to ligature":
            warnings.warn("Mark to ligature not yet supported")

        elif self.current_lookup_type == "context":
            if line.startswith("glyph"):
                m = line.rstrip().split("\t")
                context = [[x] for x in re.split(r",\s*", m[1])]
                self.add_chain_simple(context, m[2:])
            elif line.startswith("class"):
                m = line.rstrip().split("\t")
                m[1] = m[1].replace(" ", "")
                context = self.make_context(re.split(r",\s*", m[1]), self.classContexts)
                self.add_chain_simple(context, m[2:])
        elif self.current_lookup_type == "chained":
            if line.startswith("class-chain"):
                m = re.match("class-chain\t([^\t]*)\t([^\t]*)\t([^\t]*)\t(.*)$", line)
                precontext = []
                if m[1]:
                    precontext = self.make_context(
                        re.split(r",\s*", m[1]), self.backtrackclassContexts
                    )
                    precontext = list(reversed(precontext))
                context = self.make_context(re.split(r",\s*", m[2]), self.classContexts)
                postcontext = []
                if m[3]:
                    postcontext = self.make_context(
                        re.split(r",\s*", m[3]), self.lookaheadclassContexts
                    )
                lookups = m[4].rstrip().split("\t")
                # print("Lookup %s, lookups = %s" % (self.current_lookup.name, lookups))
                self.current_lookup.addRule(
                    Chaining(
                        context,
                        precontext=precontext,
                        postcontext=postcontext,
                        lookups=lookups,
                    )
                )
            elif line.startswith("glyph"):
                raise ValueError("GSUB6.1 not supported yet")
        else:
            raise ValueError("Unsupported lookup type |%s|" % self.current_lookup_type)

    def make_context(self, classlist, which):
        context = []
        for x in classlist:
            x = x.strip()
            if x == "0":
                if not self.glyphset:
                    raise ValueError(
                        "Class 0 in contextual (%s) but I don't know the glyphset"
                        % self.current_lookup.name
                    )
                else:
                    # Class 0 is everything that's not mentioned elsewhere
                    members = set(self.glyphset)
                    for c in which.values():
                        members = members - set(c)
            else:
                if not x in which:
                    print(
                        "Couldn't find a class definition for class %s in lookup %s"
                        % (x, self.current_lookup.name)
                    )
                members = which[x]
            context.append(members)
        return context

    def add_to_class_definition(self, which, line):
        m = re.match("([\w\.]+)\s+(\d+)", line)
        if not m:
            raise ValueError("Unparsable line '%s'" % line)
        if which == "class":
            which = self.classContexts
        elif which == "backtrackclass":
            which = self.backtrackclassContexts
        elif which == "lookaheadclass":
            which = self.lookaheadclassContexts

        if not m[2] in which:
            which[m[2]] = []
        which[m[2]].append(m[1])


def unparse(filename, config={}, font=None):
    if config:
        import json

        with open(config) as f:
            config = json.load(f)
    else:
        config = {}
    if font:
        glyphset = TTFont(font).getGlyphSet().keys()
    else:
        glyphset = ()
    with open(filename) as file_in:
        parser = FontDameParser(file_in, config, glyphset)
        parser.parse()
    output = ""
    done = {}

    return parser.ff



# Parse a GPOS ttx dump and output a cursive attachment .fea
import xml.etree.cElementTree as et
from fontTools.feaLib.ast import *

tree = et.parse("cursive.xml")
doc = tree.getroot()
glyphs = [x.attrib["value"] for x in tree.findall("Coverage/Glyph")]
recs = tree.findall("EntryExitRecord")

droplist = []
poslist = []
for i,rec in enumerate(recs):
  entries = rec.findall("EntryAnchor")
  exits   = rec.findall("ExitAnchor")
  if len(entries) > 0:
    entryY = int(entries[0].find("YCoordinate").attrib["value"])
    entryX = int(entries[0].find("XCoordinate").attrib["value"])
    entryAnchor = Anchor(entryX, entryY)
  else:
    entryY = 0
    entryAnchor = Anchor(0,0,name="NULL")

  if len(exits) > 0:
    exitY = int(exits[0].find("YCoordinate").attrib["value"])
    exitX = int(exits[0].find("XCoordinate").attrib["value"])
    exitAnchor = Anchor(exitX, exitY)
  else:
    exitY = 0
    exitAnchor = Anchor(0,0,name="NULL")

  drop = entryY - exitY
  droplist.append((glyphs[i],drop))
  poslist.append(CursivePosStatement(GlyphName(glyphs[i]), entryAnchor, exitAnchor))

droplist = sorted(droplist, key = lambda x:x[1])

import numpy as np

def groupByDrop(droplist, classname, bincount=10):
  relevant = list(filter(lambda x: classname in x[0] , droplist))
  dropValues = [ x[1] for x in relevant ]
  edges = np.histogram_bin_edges(dropValues, bins = bincount)
  middles = list((edges[0:-1] + edges[1:] ) /2)
  groups = np.digitize(dropValues, edges) -1

  binnedGlyphs = {}
  for i,group in enumerate(groups):
    if not group in binnedGlyphs:
      binnedGlyphs[group] = []
    binnedGlyphs[group].append(relevant[i][0])
  middles.append(edges[-1])
  classes = []
  for k,v in binnedGlyphs.items():
    classes.append(GlyphClassDefinition("%s_%i" % (classname,k), GlyphClass(v)))
    classes.append(Comment("# average drop = %i\n" % middles[k]))

  return binnedGlyphs, middles, classes

binnedMeds, medMiddles, medClasses = groupByDrop(droplist, "Med")
for x in medClasses:
  print(x.asFea())
binnedInis, iniMiddles, iniClasses = groupByDrop(droplist, "Ini")
for x in iniClasses:
  print(x.asFea())

binnedFins, finMiddles, finClasses = groupByDrop(droplist, "Fin")
for x in finClasses:
  print(x.asFea())

print("feature curs {\n")
print("lookupflag IgnoreMarks;\n")
for x in poslist:
  print(x.asFea())
print("} curs;")

# print("feature kern {\n")
# print("lookupflag IgnoreMarks;\n")

# for i in binnedInis.keys():
#   for m1 in binnedMeds.keys():
#     for f in binnedFins.keys():
#       totalDrop = (iniMiddles[i] + medMiddles[m1] + finMiddles[f])
#       print("\tposition @Ini_%i' <0 %i 0 0> @Med_%i @Fin_%i;" % (i, totalDrop, m1,f))

# for i in binnedInis.keys():
#   for m1 in binnedMeds.keys():
#     for m2 in binnedMeds.keys():
#       for f in binnedFins.keys():
#         totalDrop = (iniMiddles[i] + medMiddles[m1] + medMiddles[m2] + finMiddles[f])
#         print("\tposition @Ini_%i' <0 %i 0 0> @Med_%i  @Med_%i  @Fin_%i;" % (i, totalDrop, m1,m2,f))
#   print("} kern;")



"""A Python Unicode Shaping Engine."""

from fontFeatures import FontFeatures
import unicodedata
from fontFeatures.shaperLib import Buffer
from .BaseShaper import BaseShaper
from .ArabicShaper import ArabicShaper
from .IndicShaper import IndicShaper
from .MyanmarShaper import MyanmarShaper
from .HangulShaper import HangulShaper
from .KhmerShaper import KhmerShaper
from .USEShaper import USEShaper
import logging
import re


class Shaper:
    """Initialize a shaping engine.

    Args:
        ff: A :py:class:`fontFeatures.FontFeatures` object.
        font: A ``Babelfont`` font object.
        message_function: A function called with a message and buffer object.
    """
    def __init__(self, ff, font, message_function=None):
        assert isinstance(ff, FontFeatures)
        self.fontfeatures = ff
        self.babelfont = font
        if message_function:
            self.msg = message_function
        else:
            self.msg = self.default_message_function

    def execute(self, buf, features=[]):
        """Run the shaper on a given buffer.

        Args:
            buf: :py:class:`fontFeatures.shaperLib.buffer.Buffer` object.
            features: either a list of feature tags or a feature string ("+foox,-barx")
        """
        # Choose complex shaper
        self.complexshaper = self.categorize(buf)(self, self.babelfont, buf, features)
        self.msg("Using %s" % type(self.complexshaper).__name__)
        self.stages = [[]]
        if isinstance(features, str):
            self.user_features = self._parse_user_feature_string(features)
        else:
            self.user_features = features
        self.collect_features(buf)
        self.complexshaper.shape()
        buf.clear_mask()
        return buf

    def default_message_function(self, msg, buffer=None, serialize_options=None):
        """A logger function to be used if one is not provided in the constructor.

        By default, debugging and informational log messages are reported to the
        ``fontFeatures.shaperLib`` facility."""
        ser = ""
        if buffer:
            ser = buffer.serialize(additional=serialize_options)
            msg = msg + " : " + ser
        logging.getLogger("fontFeatures.shaperLib").info(msg)

    def _parse_user_feature_string(self, features):
        features = features.split(",")
        outfeat = []
        for f in features:
            f = f.rstrip()
            m = re.match(r"^([+\-]?)(\w+)", f)
            if m:
                outfeat.append({"tag": m[2], "value": m[1] != "-"})
                continue
            m = re.match(r"^(\w+)=([10])", f)
            if m:
                outfeat.append({"tag": m[1], "value": m[2] == "1"})
            else:
                outfeat.append({"tag": f, "value": True})
        return outfeat

    def add_pause(self, thing = None):
        """Add a pause in the shaping stage. Used internally."""
        if thing:
            self.stages.append(thing)
        self.stages.append([])

    def add_features(self, *tags):
        """Adds feature tags to the current shaping stage."""
        for t in tags:
            if any([isinstance(x, list) and t in x for x in self.stages]):
                continue
            self.stages[-1].append(t)

    def disable_feature(self, tag):
        """Removes a feature from processing."""
        for s in self.stages:
            if isinstance(s, list) and tag in s:
                s.remove(tag)

    def collect_features(self, buf):
        """Determine the features, and their order, to process the buffer."""
        self.add_features("rvrn")
        self.add_pause()
        if buf.direction == "LTR":
            self.add_features("ltra", "ltrm")
        elif buf.direction == "RTL":
            self.add_features("rtla", "rtlm")
        self.add_features("frac", "numr", "dnom", "rand")
        # trak?
        self.complexshaper.collect_features(self)
        # common features
        self.add_features("abvm", "blwm", "ccmp", "locl", "mark", "mkmk", "rlig")
        if buf.direction == "LTR" or buf.direction == "RTL":
            self.add_features("calt", "clig", "curs", "dist", "kern", "liga", "rclt")
        else:
            self.add_features("vert")
        for uf in self.user_features:
            if not uf["value"]:  # Turn it off if it's already on
                self.disable_feature(uf["tag"])
            else:
                self.add_features(uf["tag"])
        if hasattr(self.complexshaper, "override_features"):
            self.complexshaper.override_features(self)

    def categorize(self, buf):
        """Returns the appropriate complex shaper class to shape this buffer."""
        if buf.script == "Arabic":
            return ArabicShaper

        connected_scripts = {
            "Mongolian": "mong",
            "Syriac": "syrc",
            "Nko": "nko ",
            "Phags_Pa": "phag",
            "Mandaic": "mand",
            "Manichaean": "mand",
            "Psalter_Pahlavi": "phlp",
            "Adlam": "adlm",
            "Hanifi_Rohingya": "rohg",
            "Sogdian": "sogd"
        }
        if buf.script in connected_scripts:
            if self.fontfeatures.hasScriptSupport(connected_scripts[buf.script]):
                return ArabicShaper
            else:
                return BaseShaper

        if buf.script in ["Thai", "Lao"]:
            return ThaiShaper
        if buf.script == "Hangul":
            return HangulShaper
        if buf.script == "Hebrew":
            return HebrewShaper
        if buf.script in [
            "Bengali",
            "Devanagari",
            "Gujarati",
            "Gurmukhi",
            "Kannada",
            "Malayalam",
            "Oriya",
            "Tamil",
            "Telugu",
            "Sinhala",
        ]:
            indic23map = {
                "Bengali": "bng",
                "Devanagari": "dev",
                "Gujarati": "gjr",
                "Gurmukhi": "gur",
                "Kannada": "knd",
                "Malayalam": "mlm",
                "Oriya": "ory",
                "Tamil": "tml",
                "Telugu": "tel",
            }
            # Sinhala is different
            if buf.script in indic23map and self.fontfeatures.hasScriptSupport(indic23map[buf.script] + "3"):
                return USEShaper
            else:
                return IndicShaper
        if buf.script == "Khmer":
            return KhmerShaper

        if buf.script == "Myanmar":
            if self.fontfeatures.hasScriptSupport("mymr"):
                return BaseShaper
            else:
                return MyanmarShaper

        # if buf.script = "Qaag": return MyanmarZawgyiShaper
        if buf.script in [
            "Tibetan",
            "Buhid",
            "Hanunoo",
            "Tagalog",
            "Tagbanwa",
            "Limbu",
            "Tai_Le",
            "Buginese",
            "Kharoshthi",
            "Syloti_Nagri",
            "Tifinagh",
            "Balinese",
            "Cham",
            "Kayah_Li",
            "Lepcha",
            "Rejang",
            "Saurashtra",
            "Sundanese",
            "Egyptian_Hieroglyphs",
            "Javanese",
            "Kaithi",
            "Meetei_Mayek",
            "Tai_Tham",
            "Tai_Viet",
            "Batak",
            "Brahmi",
            "Chakma",
            "Sharada",
            "Takri",
            "Duployan",
            "Grantha",
            "Khojki",
            "Khudawadi",
            "Mahajani",
            "Modi",
            "Pahawh_Hmong",
            "Siddham",
            "Tirhuta",
            "Ahom",
            "Bhaiksuki",
            "Marchen",
            "Newa",
            "Masaram_Gondi",
            "Soyombo",
            "Zanabazar_Square",
            "Dogra",
            "Gunjala_Gondi",
            "Makasar",
            "Nandinagari",
        ]:
            return USEShaper
        return BaseShaper

def _script_direction(script):
    if script in [
        "Arabic",
        "Hebrew",
        "Syriac",
        "Thaana",
        "Cypriot",
        "Kharoshthi",
        "Phoenician",
        "Nko",
        "Lydian",
        "Avestan",
        "Imperial_Aramaic",
        "Inscriptional_Parthian",
        "Inscriptional_Pahlavi",
        "Old_South_Arabian",
        "Old_Turkic",
        "Samaritan",
        "Mandaic",
        "Meroitic_Cursive",
        "Meroitic_Hieroglyphs",
        "Manichaean",
        "Mende_Kikakui",
        "Nabataean",
        "Old_North_Arabian",
        "Palmyrene",
        "Psalter_Pahlavi",
        "Hatran",
        "Adlam",
        "Hanifi_Rohingya",
        "Old_Sogdian",
        "Sogdian",
        "Elymaic",
        "Chorasmian",
        "Yezidi"]:
        return "RTL"
    if script in ["Old_Hungarian", "Old_Italic", "Runic"]:
        return "invalid"
    return "LTR"



from .BaseShaper import BaseShaper
from youseedee import ucd_data


jts = {
  "U": 0, "L": 1, "R": 2, "D": 3, "ALAPH": 4, "DALATH_RISH": 5
}
state_table = [
  #   jt_U,          jt_L,          jt_R,          jt_D,          jg_ALAPH,      jg_DALATH_RISH */

  # State 0: prev was U, not willing to join.
  [ ("none","none",0), ("none","isol",2), ("none","isol",1), ("none","isol",2), ("none","isol",1), ("none","isol",6), ],

  # State 1: prev was R or ISOL/ALAPH, not willing to join.
  [ ("none","none",0), ("none","isol",2), ("none","isol",1), ("none","isol",2), ("none","fin2",5), ("none","isol",6), ],

  # State 2: prev was D/L in ISOL form, willing to join.
  [ ("none","none",0), ("none","isol",2), ("init","fina",1), ("init","fina",3), ("init","fina",4), ("init","fina",6), ],

  # State 3: prev was D in FINA form, willing to join.
  [ ("none","none",0), ("none","isol",2), ("medi","fina",1), ("medi","fina",3), ("medi","fina",4), ("medi","fina",6), ],

  # State 4: prev was FINA ALAPH, not willing to join.
  [ ("none","none",0), ("none","isol",2), ("med2","isol",1), ("med2","isol",2), ("med2","fin2",5), ("med2","isol",6), ],

  # State 5: prev was FIN2/FIN3 ALAPH, not willing to join.
  [ ("none","none",0), ("none","isol",2), ("isol","isol",1), ("isol","isol",2), ("isol","fin2",5), ("isol","isol",6), ],

  # State 6: prev was DALATH/RISH, not willing to join.
  [ ("none","none",0), ("none","isol",2), ("none","isol",1), ("none","isol",2), ("none","fin3",5), ("none","isol",6),],
]

arabic_features = ['isol', 'fina', 'fin2', 'fin3', 'medi', 'med2', 'init']

class ArabicShaper(BaseShaper):
    def collect_features(self, shaper):
        # shaper.add_features("stch")
        shaper.add_features("ccmp", "locl")
        shaper.add_pause()
        shaper.add_features(*arabic_features)
        shaper.add_pause()
        shaper.add_features("rlig", "rclt", "calt")
        shaper.add_pause()
        shaper.add_features("mset")

    def substitute_default(self):
        super().substitute_default()
        state = 0
        prev_item = None
        for item in self.buffer.items:
            item.arabic_joining = "NONE"
            ucd = ucd_data(item.codepoint)
            joining = ucd.get("Joining_Type")
            if not joining:
                if ucd.get("General_Category") in ["Mn", "Cf", "Em"]:
                    joining = "T"
                else:
                    joining = "U"
            if joining == "T": continue
            if joining == "C": joining = "D"  # Mongolian
            if ucd.get("Joining_Group") == "ALAPH": joining = "ALAPH"
            if ucd.get("Joining_Group") == "DALATH RISH": joining = "DALATH_RISH"
            prev, this, state = state_table[state][jts[joining]]
            if prev_item:
              prev_item.arabic_joining = prev
            item.arabic_joining = this
            prev_item = item
        if self.buffer.script == "Mongolian":
            self.mongolian_variation_selectors()
        self.plan.msg("Assigned Arabic joining", self.buffer, serialize_options=["arabic_joining"])
        for f in arabic_features:
            if f not in self.plan.fontfeatures.features:
                continue
            for item in self.buffer.items:
                item.feature_masks[f] = item.arabic_joining != f

    def mongolian_variation_selectors(self):
        for ix,item in enumerate(self.buffer.items):
            if 0x180B <= item.codepoint <= 0x180D and ix > 0:
                item.arabic_joining = self.buffer.items[ix-1].arabic_joining



"""The base class for complex shapers.

Shaper.py is the top-level interface to shaperlib. This is a lower-level
module implementing common routines used to the individual complex shapers."""

from fontFeatures.shaperLib.Buffer import Buffer, BufferItem
from copy import copy
import unicodedata
from youseedee import ucd_data
from fontFeatures import RoutineReference



def _is_default_ignorable(c):
    return (
        c in [0x00AD, 0x034F, 0x061C, 0x17B4, 0x17B5, 0xFEFF]
        or 0x180B <= c <= 0x180E
        or 0x200B <= c <= 0x200F
        or 0x202A <= c <= 0x202E
        or 0x2060 <= c <= 0x206F
        or 0xFE00 <= c <= 0xFE0F
        or 0xFFF0 <= c <= 0xFFF8
        or 0x1D173 <= c <= 0x1D17A
        or 0xE0000 <= c <= 0xE0FFF
    )


class BaseShaper:
    """A complex shaper object."""
    def __init__(self, plan, font, buf, features=[]):
        self.plan = plan
        self.font = font
        self.buffer = buf
        self.features = features

    def shape(self):
        """Shape a buffer."""
        self.plan.fontfeatures.resolveAllRoutines()
        # self.buffer.set_unicode_props()
        # self.insert_dotted_circles()
        # self.buffer.form_clusters()
        # self.buffer.ensure_native_direction()
        if not self.buffer.is_all_glyphs:
            self.preprocess_text()
            # Substitute pre
            self.substitute_default()
        self.substitute_complex()
        # Substitute post
        self.delete_default_ignorables()

        self.position()
        self.postprocess_glyphs()
        # self.buffer.propagate_flags()

    def preprocess_text(self):
        """Do any processing which needs to happen to character items before shaping."""
        pass

    def postprocess_glyphs(self):
        """Do any processing which needs to happen to glyph items after shaping."""
        pass

    def substitute_default(self):
        """Do initial subsitution processing: buffer normalization and mapping."""
        self.normalize_unicode_buffer()
        self.buffer.map_to_glyphs()
        self.plan.msg("Initial glyph mapping", self.buffer)
        # Setup masks
        # if self.buf.fallback_mark_positioning:
        # self.fallback_mark_position_recategorize_marks()
        pass

    def normalize_unicode_buffer(self):
        """Normalize an OpenType buffer."""
        unistring = "".join([chr(item.codepoint) for item in self.buffer.items])
        self.buffer.store_unicode(unicodedata.normalize("NFC", unistring))

        # Some fix-ups from hb-ot-shape-normalize
        for item in self.buffer.items:
            if ucd_data(item.codepoint)[
                "General_Category"
            ] == "Zs" and self.font.unicode_map.get(0x20, None):
                item.codepoint = 0x20
                # Harfbuzz adjusts the width here, in _hb_ot_shape_fallback_spaces
            if item.codepoint == 0x2011 and self.font.unicode_map.get(0x2010, None):
                item.codepoint = 0x2010

    def collect_features(self, shaper):
        """Collect all complex features to be run by this shaper. (Abstract method.)"""
        return []

    def substitute_complex(self):
        """Run the substitution stage."""
        self._run_stage("sub")

    def position(self):
        """Run the positioning stage."""
        self._run_stage("pos")
        self.buffer.clear_mask()
        # zero width marks
        for i in self.buffer.items:
            if i.category[0] == "mark":
                i.position.xAdvance = 0
        # zero width default ignorables
        self.zero_width_default_ignorables()
        for i in range(0,len(self.buffer.items)):
            self.propagate_attachment_offsets(i)
        self.plan.msg("Positioning done", self.buffer)

    def propagate_attachment_offsets(self, i):
        """Resolve attachment offsets for item ``i``."""
        if not hasattr(self.buffer.items[i], "attach_type"):
            return
        attach_type = self.buffer.items[i].attach_type
        attach_chain = self.buffer.items[i].attach_chain
        if not attach_chain:
            return
        self.buffer.items[i].attach_chain = None
        j = i + attach_chain
        if j >= len(self.buffer.items):
            return
        self.propagate_attachment_offsets(j)
        if attach_type == "cursive":
            pass
            # self.buffer.items[i].position.yPlacement = (self.buffer.items[i].position.yPlacement or 0) + (self.buffer.items[j].position.yPlacement or 0) # XXX Horizontal only
        else:
            self.buffer.items[i].position.xPlacement += self.buffer.items[j].position.xPlacement or 0
            self.buffer.items[i].position.yPlacement += self.buffer.items[j].position.yPlacement or 0
            assert j < i
            if self.buffer.direction == "LTR":
                for k in range(j,i):
                    self.buffer.items[i].position.xPlacement -= self.buffer.items[k].position.xAdvance
                    self.buffer.items[i].position.yPlacement -= self.buffer.items[k].position.yAdvance or 0
            else:
                for k in range(j+1,i+1):
                    self.buffer.items[i].position.xPlacement += self.buffer.items[k].position.xAdvance
                    self.buffer.items[i].position.yPlacement += self.buffer.items[k].position.yAdvance or 0


    def _run_stage(self, current_stage):
        self.plan.msg("Running %s stage" % current_stage)
        self.plan.fontfeatures.hoist_languages()
        for stage in self.plan.stages:
            lookups = []
            if isinstance(stage, list):  # Features
                for f in stage:
                    if f not in self.plan.fontfeatures.features:
                        continue

                    routines = self.plan.fontfeatures.features[f]
                    routines = [x.routine if isinstance(x, RoutineReference) else x for x in routines]
                    lookups.extend(
                        [(routine, f) for routine in self._filter_by_lang(routines)]
                    )
                self.plan.msg("Processing features: %s" % ",".join(stage))
                for r, feature in lookups:
                    self.plan.msg(
                        "Before %s (%s)" % (r.name, feature), buffer=self.buffer
                    )
                    r.apply_to_buffer(self.buffer, stage=current_stage, feature=feature, namedclasses=self.plan.fontfeatures.namedClasses)
                    self.plan.msg(
                        "After %s (%s)" % (r.name, feature), buffer=self.buffer
                    )
            else:
                # It's a pause. We only support GSUB pauses.
                if current_stage == "sub":
                    stage(current_stage)

    def _filter_by_lang(self, routines):
        script = self.script_to_opentype.get(self.buffer.script,"DFLT")
        s_l = self.plan.fontfeatures.scripts_and_languages
        if not self.buffer.script or len(s_l.values()) < 2 and (
            len(s_l.values()) == 0 or len(list(s_l.values())[0]) < 2):
            return routines # !
        language = self.buffer.language or "dflt"
        if script in s_l and (script, language) != ("DFLT", "dflt"):
            return [x for x in routines if x.languages and (script, language) in x.languages]
        else:
            return [x for x in routines if not x.languages or ("DFLT", language) in x.languages]

    def delete_default_ignorables(self):
        """Remove all items from the buffer which are ignorable."""
        self.buffer.items = [x for x in self.buffer.items if not(x.codepoint) or not _is_default_ignorable(x.codepoint)]

    def zero_width_default_ignorables(self):
        """Set all items from the buffer which are ignorable to zero advance width."""
        space = BufferItem.new_unicode(0x20)
        space.map_to_glyph(self.buffer.font)
        if space.glyph == -1:
            return
        for ix, item in enumerate(self.buffer.items):
            if item.codepoint and _is_default_ignorable(item.codepoint):
                item.glyph = space.glyph
                item.position.xAdvance = 0
                for i in self.buffer.items[ix:]:
                    if hasattr(i, "syllable_index"):
                        i.syllable_index += 1

    def would_substitute(self, feature, subbuffer_items):
        """Test whether the feature would apply to a list of BufferItem objects."""
        if not feature in self.plan.fontfeatures.features:
            return False
        subbuffer = Buffer(
            self.buffer.font,
            direction=self.buffer.direction,
            script=self.buffer.script,
            language=self.buffer.language,
        )
        subbuffer.clear_mask()
        subbuffer.items = [copy(x) for x in subbuffer_items]
        subbuffer.clear_mask()
        routines = self.plan.fontfeatures.features[feature]
        for r in routines:
            if isinstance(r, RoutineReference):
                r = r.routine
            for rule in r.rules:
                if rule.stage == "pos":
                    continue
                if any(
                    [
                        rule.would_apply_at_position(subbuffer, i, namedclasses=self.plan.fontfeatures.namedClasses)
                        for i in range(len(subbuffer))
                    ]
                ):
                    return True
        return False

    script_to_opentype = {
        "Common": "zyyy",
        "Inherited": "zinh",
        "Unknown": "zzzz",
        "Arabic": "arab",
        "Armenian": "armn",
        "Bengali": "bng2", # Not beng
        "Cyrillic": "cyrl",
        "Devanagari": "dev2", # Not deva
        "Georgian": "geor",
        "Greek": "grek",
        "Gujarati": "gjr2", # Not gujr
        "Gurmukhi": "gur2", # Not guru
        "Hangul": "hang",
        "Han": "hani",
        "Hebrew": "hebr",
        "Hiragana": "hira",
        "Kannada": "knd2", # Not knda
        "Katakana": "kana",
        "Lao": "laoo",
        "Latin": "latn",
        "Malayalam": "mlm2", # Not mlym
        "Oriya": "ory2", # Not orya
        "Tamil": "tml2", # Not taml
        "Telugu": "tel2", # Not telu
        "Thai": "thai",
        "Tibetan": "tibt",
        "Bopomofo": "bopo",
        "Braille": "brai",
        "Canadian_Syllabics": "cans",
        "Cherokee": "cher",
        "Ethiopic": "ethi",
        "Khmer": "khmr",
        "Mongolian": "mong",
        "Myanmar": "mym2", # Not mymr
        "Ogham": "ogam",
        "Runic": "runr",
        "Sinhala": "sinh",
        "Syriac": "syrc",
        "Thaana": "thaa",
        "Yi": "yiii",
        "Deseret": "dsrt",
        "Gothic": "goth",
        "Old_Italic": "ital",
        "Buhid": "buhd",
        "Hanunoo": "hano",
        "Tagalog": "tglg",
        "Tagbanwa": "tagb",
        "Cypriot": "cprt",
        "Limbu": "limb",
        "Linear_B": "linb",
        "Osmanya": "osma",
        "Shavian": "shaw",
        "Tai_Le": "tale",
        "Ugaritic": "ugar",
        "Buginese": "bugi",
        "Coptic": "copt",
        "Glagolitic": "glag",
        "Kharoshthi": "khar",
        "New_Tai_Lue": "talu",
        "Old_Persian": "xpeo",
        "Syloti_Nagri": "sylo",
        "Tifinagh": "tfng",
        "Balinese": "bali",
        "Cuneiform": "xsux",
        "Nko": "nkoo",
        "Phags_Pa": "phag",
        "Phoenician": "phnx",
        "Carian": "cari",
        "Cham": "cham",
        "Kayah_Li": "kali",
        "Lepcha": "lepc",
        "Lycian": "lyci",
        "Lydian": "lydi",
        "Ol_Chiki": "olck",
        "Rejang": "rjng",
        "Saurashtra": "saur",
        "Sundanese": "sund",
        "Vai": "vaii",
        "Avestan": "avst",
        "Bamum": "bamu",
        "Egyptian_Hieroglyphs": "egyp",
        "Imperial_Aramaic": "armi",
        "Inscriptional_Pahlavi": "phli",
        "Inscriptional_Parthian": "prti",
        "Javanese": "java",
        "Kaithi": "kthi",
        "Lisu": "lisu",
        "Meetei_Mayek": "mtei",
        "Old_South_Arabian": "sarb",
        "Old_Turkic": "orkh",
        "Samaritan": "samr",
        "Tai_Tham": "lana",
        "Tai_Viet": "tavt",
        "Batak": "batk",
        "Brahmi": "brah",
        "Mandaic": "mand",
        "Chakma": "cakm",
        "Meroitic_Cursive": "merc",
        "Meroitic_Hieroglyphs": "mero",
        "Miao": "plrd",
        "Sharada": "shrd",
        "Sora_Sompeng": "sora",
        "Takri": "takr",
        "Bassa_Vah": "bass",
        "Caucasian_Albanian": "aghb",
        "Duployan": "dupl",
        "Elbasan": "elba",
        "Grantha": "gran",
        "Khojki": "khoj",
        "Khudawadi": "sind",
        "Linear_A": "lina",
        "Mahajani": "mahj",
        "Manichaean": "mani",
        "Mende_Kikakui": "mend",
        "Modi": "modi",
        "Mro": "mroo",
        "Nabataean": "nbat",
        "Old_North_Arabian": "narb",
        "Old_Permic": "perm",
        "Pahawh_Hmong": "hmng",
        "Palmyrene": "palm",
        "Pau_Cin_Hau": "pauc",
        "Psalter_Pahlavi": "phlp",
        "Siddham": "sidd",
        "Tirhuta": "tirh",
        "Warang_Citi": "wara",
        "Ahom": "ahom",
        "Anatolian_Hieroglyphs": "hluw",
        "Hatran": "hatr",
        "Multani": "mult",
        "Old_Hungarian": "hung",
        "Signwriting": "sgnw",
        "Adlam": "adlm",
        "Bhaiksuki": "bhks",
        "Marchen": "marc",
        "Osage": "osge",
        "Tangut": "tang",
        "Newa": "newa",
        "Masaram_Gondi": "gonm",
        "Nushu": "nshu",
        "Soyombo": "soyo",
        "Zanabazar_Square": "zanb",
        "Dogra": "dogr",
        "Gunjala_Gondi": "gong",
        "Hanifi_Rohingya": "rohg",
        "Makasar": "maka",
        "Medefaidrin": "medf",
        "Old_Sogdian": "sogo",
        "Sogdian": "sogd",
        "Elymaic": "elym",
        "Nandinagari": "nand",
        "Nyiakeng_Puachue_Hmong": "hmnp",
        "Wancho": "wcho",
        "Chorasmian": "chrs",
        "Dives_Akuru": "diak",
        "Khitan_Small_Script": "kits",
        "Yezidi": "yezi",
    }




# Code for converting a Substitution object into feaLib statements
import fontTools.feaLib.ast as feaast
from fontFeatures.ttLib.Substitution import lookup_type
from itertools import cycle
from typing import Union

def glyphref(g) -> Union[feaast.GlyphName, feaast.GlyphClass]:
    if len(g) == 1:
        return feaast.GlyphName(g[0])
    return feaast.GlyphClass([feaast.GlyphName(x) for x in g])

# Whether or not the input or the replacement has classes (so, entire statement).
def has_classes(self) -> bool:
    input_has_class = next((i for i,v in enumerate([len(x) for x in self.input]) if v>1), None)
    replacement_has_class = next((i for i,v in enumerate([len(x) for x in self.replacement]) if v>1), None)

    return (input_has_class is not None or replacement_has_class is not None)

# Excepting single glyphs, if classes are defined, the input and replacement
# all use classes of equal length.
def all_classes_equal(self) -> bool:
    input_lengths = [len(x) for x in self.input if len(x) != 1]
    replacement_lengths = [len(x) for x in self.replacement if len(x) != 1]

    if len(input_lengths) == 0 and len(replacement_lengths) == 0:
        return True

    return len(set(input_lengths+replacement_lengths)) == 1

# One of the substitution/replacements has all-but-one arity one,
# and both arities are the same
def is_paired(self) -> bool:
    input_lengths = [len(x) for x in self.input if len(x) != 1]
    replacement_lengths = [len(x) for x in self.replacement if len(x) != 1]
    if not (len(input_lengths) == 1 and len(replacement_lengths) == 1):
        return False
    if input_lengths[0] != replacement_lengths[0]:
        import warnings
        warnings.warn("Unbalanced paired substitution")
        return False
    return True

# Expand ligature substitutions ending in a class, such that:
#   * Substitute [f f.ss01] i -> [f_i f_i.ss01];
# Expands to:
#   * sub f i -> f_i;
#   * sub f.ss01 i -> f_i.ss01;
# Likewise:
#   * Substitute [f f.ss01] [i i.ss01] -> [f_i f_i.ss01];
# Expands to:
#   * sub f i -> f_i;
#   * sub f.ss01 i.ss01 -> f_i.ss01;
def paired_ligature(self) -> feaast.LigatureSubstStatement:
    b = feaast.Block()
    inputs = []
    for i in self.input:
        if len(i) == 1:
            inputs.append(cycle(i))
        else:
            inputs.append(i)
    lhs = zip(*inputs)
    replacements = []
    for j in self.replacement:
        if len(j) == 1:
            replacements.append(cycle(j))
        else:
            replacements.append(j)
    rhs = zip(*replacements)

    for l, r in zip(lhs,rhs):
        stmt = feaast.LigatureSubstStatement(
            [glyphref(x) for x in self.precontext],
            [glyphref([x]) for x in l],
            [glyphref(x) for x in self.postcontext],
            glyphref([r[0]]),
            False,
        )
        b.statements.append(stmt)
    return b

# Expand multiple substitutions, such that:
#   * Substitute [a b] -> before_tail [a.2 b.2] tail;
# Becomes in FEA:
#   * sub a by before_tail a.2 tail;
#   * sub b by before_tail b.2 tail;
def paired_mult(self) -> feaast.MultipleSubstStatement:
    b = feaast.Block()

    input_lengths = [len(x) for x in self.input]
    replacement_lengths = [len(x) for x in self.replacement]

    if len(input_lengths) != 1:
        raise ValueError("Multiple substitution only valid on input of length one, use a Chain instead")

    input_length = input_lengths[0]

    if not sum([l for l in replacement_lengths if l == 1]) in [len(replacement_lengths), len(replacement_lengths)-1]:
        raise ValueError("Cannot expand multiple glyph classes in a multiple substitution — creates ambiguity")

    # Look for the glyph class in the replacement, or default to first glyph in replacement
    glyphcls = next((i for i, v in enumerate(self.replacement) if len(v)> 1), 0)

    if input_length != len(self.replacement[glyphcls]):
        raise ValueError("Glyph class in input must be same length as that in replacement. {} != {}".format(input_length, len(self.replacement[glyphcls])))

    zipped = zip(self.input[0], self.replacement[glyphcls])

    prior_reps = self.replacement[:glyphcls]
    after_reps = self.replacement[glyphcls+1:]

    for f, t in zipped:
        stmt = feaast.MultipleSubstStatement(
            [glyphref(x) for x in self.precontext],
            glyphref([f]),
            [glyphref(x) for x in self.postcontext],
            [glyphref(g) for g in prior_reps+[[t]]+after_reps]
        )
        b.statements.append(stmt)

    return b

def asFeaAST(self):
    lut = lookup_type(self, forFea=True)
    if not lut:
        return feaast.Comment("")

    if not self.replacement: # Delete
        return feaast.MultipleSubstStatement(
            [glyphref(x) for x in self.precontext],
            glyphref(self.input[0]),
            [glyphref(x) for x in self.postcontext],
            [],
        )

    if lut == 1: # GSUB 1 Single Substitution
        return feaast.SingleSubstStatement(
            [glyphref(x) for x in self.input],
            [glyphref(x) for x in self.replacement],
            [glyphref(x) for x in self.precontext],
            [glyphref(x) for x in self.postcontext],
            False,
        )
    elif lut == 2: # GSUB 2 Multiple Substitution
        # Paired rules need to become a set of statements
        if is_paired(self):
            return paired_mult(self)

        return feaast.MultipleSubstStatement(
            [glyphref(x) for x in self.precontext],
            glyphref(self.input[0]),
            [glyphref(x) for x in self.postcontext],
            [glyphref(x) for x in self.replacement],
        )
    elif lut == 3: # GSUB 3 Alternate Substitution
        return feaast.AlternateSubstStatement(
            [glyphref(x) for x in self.precontext],
            glyphref(self.input[0]),
            [glyphref(x) for x in self.postcontext],
            feaast.GlyphClass([feaast.GlyphName(x) for x in self.replacement[0]]),
        )
    elif lut == 4: # GSUB 4 Ligature Substitution
        # Some rules with classes need to become a set of statements.
        if has_classes(self) and all_classes_equal(self) and not len(self.replacement[0]) == 1:
            return paired_ligature(self)

        return feaast.LigatureSubstStatement(
            [glyphref(x) for x in self.precontext],
            [glyphref(x) for x in self.input],
            [glyphref(x) for x in self.postcontext],
            glyphref(self.replacement[0]),
            False,
        )
    elif lut in [5, 6, 7]: # GSUB 5, 6, 7 Different types of contextual substitutions
        raise NotImplementedError("Use the Chain verb for this")
    elif lut == 8: # GSUB 8 Reverse Chaining Single Substitution
        return feaast.ReverseChainSingleSubstStatement(
            [glyphref(x) for x in self.precontext],
            [glyphref(x) for x in self.postcontext],
            [glyphref(x) for x in self.input],
            [glyphref(self.replacement[0])],
        )
    elif lut >= 9:
        raise NotImplementedError("Invalid GSUB lookup type requested: {}".format(lut))

    raise ValueError("LookupType must be a single positive integer")



# Code for converting a FontFeatures object into feaLib statements
import fontTools.feaLib.ast as feaast
import itertools
from fontFeatures.feaLib.Routine import lookup_type
import copy


def add_language_system_statements(self, ff):
    self.hoist_languages()

    total_languages = sum([len(x) for x in self.scripts_and_languages.values()])

    if total_languages < 2:
        return
    for s, entry in self.scripts_and_languages.items():
        for l in entry:
            ff.statements.append(feaast.LanguageSystemStatement(s, l))


def asFea(self,**kwargs):
    return self.asFeaAST(**kwargs).asFea()


def _to_inline_class(glyphs):
    return feaast.GlyphClass([feaast.GlyphName(x) for x in glyphs])


def reorderAndResolve(self):
    from fontFeatures import Chaining

    # Arrange the routines based on dependencies

    # First pass ensures all are references and resolves them. We do this in
    # two passes because if any bare Routines got in there, they're going to
    # to be added to the end of the routine list.
    self.resolveAllRoutines()

    # Second pass reorders
    ordering = list(range(0, len(self.routines)))
    ptr = len(self.routines) - 1
    while ptr >= 0:
        routine = self.routines[ptr]
        if not any(isinstance(r, Chaining) for r in routine.rules):
            ptr = ptr - 1
            continue

        later = []
        for r in routine.rules:
            for lookuplist in r.lookups:
                if not lookuplist:
                    continue
                for lookup in lookuplist:
                    lookup.routine.usecount = (
                        999  # Always explicitly list referenced lookups
                    )
                    ix = self.routines.index(lookup.routine)
                    if ix > ptr:
                        later.append(ix)
        for ix in later:
            f = ordering.pop(ordering.index(ix))
            ordering.insert(ptr, f)
        ptr = ptr - 1
    return ordering


def asFeaAST(self, do_gdef=True):
    """Returns this font's features as a feaLib AST object, for later
    translation to AFDKO code."""
    from fontFeatures import Routine, Chaining
    ff = feaast.FeatureFile()

    add_language_system_statements(self, ff)

    if do_gdef:
        add_gdef(self, ff)

    # In OpenType, we need to rearrange the routines such that all rules for
    # a given languagesystem, lookup type, and lookup flags, appear in the
    # same lookup. Also, lookups with the same languagesystem need to appear next
    # to one another, because FEA syntax is stupid.

    # Now arrange them by type/etc.
    for k,v in self.features.items():
        for reference in v:
            routine = reference.routine
            # If a rule has >1 language it must first be split
            newrules = []
            for r in routine.rules:
                if len(r.languages or []) > 1:
                    for language in r.languages:
                        newrule = copy.copy(r)
                        newrule.languages = [language]
                        newrules.append(newrule)
                else:
                    newrules.append(r)
            routine.rules = newrules
            partitioned = self.partitionRoutine(routine,
                lambda rule:
                    tuple([tuple(rule.languages or []),
                    type(rule),
                    lookup_type(rule)
                    ])
            )
            if routine.name and partitioned and len(partitioned) > 1:
                for p in partitioned:
                    rule = p.rules[0]
                    language = (rule.languages or [("DFLT", "dflt")])[0]
                    p.name = p.name + "%s_%s_%s_%i" % (
                        language[0].strip(), language[1].strip(), type(rule).__name__, lookup_type(rule)
                    )

    for r in self.routines:
        r.usecount = 0
        # Bubble up flags and languages
        if r.rules and not r.flags:
            r.flags = r.rules[0].flags
        if r.rules and not r.languages:
            r.languages = r.rules[0].languages

    for k, v in self.features.items():
        # Similarly split routines with multiple languages
        references = []
        for reference in v:
            routine = reference.routine
            if len(routine.languages or []) > 1:
                splitroutines = []
                for language in routine.languages:
                    # This is wrong. It should really be a new reference to the
                    # same routine. But this will do for now.
                    newreference = copy.copy(reference)
                    newreference.languages = [language]
                    references.append(newreference)
            else:
                references.append(reference)
                reference.languages = routine.languages
        self.features[k] = references
        # Order the arranged routines by language
        # new_references = list(sorted(v, key=lambda x: tuple(x.routine.languages or [])))

    # Next, we'll ensure that all chaining lookups are resolved and in the right order
    newRoutines = [self.routines[i] for i in reorderAndResolve(self)]


    # Preamble
    for k in newRoutines:
        assert isinstance(k, Routine)
        if not k.name and k.usecount != 1:
            k.name = self.gensym("Routine_")
        pre = k.feaPreamble(self)
        if k.rules:
            ff.statements.extend(pre)

    for k, v in self.namedClasses.items():
        asclass = _to_inline_class(v)
        ff.statements.append(feaast.GlyphClassDefinition(k, asclass))

    ff.statements.append(feaast.Comment(""))

    for k in newRoutines:
        if k.rules:
            ff.statements.append(k.asFeaAST())

    for k, v in self.features.items():
        for routine in v:
            # Putting each routine in its own feature saves problems...
            lang = routine.languages
            f = feaast.FeatureBlock(k)
            if lang:
                f.statements.append(feaast.ScriptStatement(lang[0][0]))
                f.statements.append(feaast.LanguageStatement("%4s" % lang[0][1]))
            f.statements.append(routine.asFeaAST(expand = k=="aalt"))
            ff.statements.append(f)
    return ff


def add_gdef(self, ff):
    gdef = feaast.TableBlock("GDEF")
    gc = self.glyphclasses
    categories = {"base": [], "mark": [], "ligature": [], "component": []}
    for k, v in gc.items():
        categories[v] = categories.get(v, []) + [k]
    if categories:
        gdef.statements.append(
            feaast.GlyphClassDefStatement(
                _to_inline_class(categories["base"]),
                _to_inline_class(categories["mark"]),
                _to_inline_class(categories["ligature"]),
                _to_inline_class(categories["component"]),
            )
        )
        ff.statements.append(gdef)



from fontFeatures import FontFeatures
from fontFeatures.shaperLib.Buffer import Buffer
from fontFeatures.shaperLib.Shaper import Shaper
from fontFeatures.ttLib import unparse
from fontTools.ttLib import TTFont
from babelfont import load
import pytest
import glob
import os
import re
import logging

if "CI" in os.environ:
    pytest.skip("Not running shaping tests on CI", allow_module_level=True)

script_map = {
    "Qaag": "Myanmar_Zawgyi"
}

# logging.getLogger("fontFeatures.shaperLib").setLevel(logging.DEBUG)

def tounicode(s):
    out = ""
    for part in s.split(","):
        if part.startswith("U+"):
            part = part[2:]
        out = out + chr(int(part,16))
    return out

tests = []
for testfile in glob.glob("tests/harfbuzz/*/tests/*.tests"):
    filetests = open(testfile,"r").readlines()
    for ix, t in enumerate(filetests):
        t = t.rstrip()
        if not t:
            continue
        if t[0] == "#":
            continue
            # t = t[1:]
        testname = testfile[:-6] + "_" + str(ix) # ".tests"
        font, hb_args, buf, expectation = t.split(":")
        font = os.path.join(os.path.dirname(testfile), font)
        if not os.path.exists(font):
            continue
        tests.append(pytest.param(font,hb_args,buf,expectation, id=testname))


@pytest.mark.parametrize("fontname,hb_args,input_string,expectation", tests)
def test_shaping(request, fontname, hb_args, input_string, expectation):
    font = TTFont(fontname)
    ff = unparse(font)
    bbf = load(fontname)
    if not bbf:
        return pytest.skip("Font too busted to use")
    buf = Buffer(bbf, unicodes=tounicode(input_string))
    shaper = Shaper(ff, bbf)
    feature_string = ""
    if "[" in hb_args:
        return pytest.skip("Partial application not supported yet")

    if "variations" in hb_args:
        return pytest.skip("Not planning to support variations")
    if "ttb" in hb_args:
        return pytest.skip("Not planning to support top-to-bottom")

    if "rand" in request.node.name:
        return pytest.skip("Not planning to support rand feature")
    if "aat" in request.node.name:
        return pytest.skip("Not planning to support Apple Advanced Typography")
    if "kern-format2" in request.node.name:
        return pytest.skip("Not planning to support kern table")
    if "fraction" in request.node.name:
        return pytest.skip("Not planning to support automatic fractions")

    m = re.search(r'--features="([^"]+)"', hb_args)
    if m:
        feature_string = m[1]

    if "--script" in hb_args:
        m = re.search(r"--script=(\w+)", hb_args)
        buf.script = script_map[m[1]]

    shaper.execute(buf, features=feature_string)
    serialize_options = {}
    if "--no-glyph-names" in hb_args:
        serialize_options["names"] = False
    if "--ned" in hb_args:
        serialize_options["ned"] = True
    if "--no-position" in hb_args:
        serialize_options["position"] = False
    if "post" not in font or font["post"].formatType != 2.0:
        serialize_options["names"] = False
        expectation = re.sub("gid", "", expectation)
    serialized = buf.serialize(**serialize_options)

    # Finesse cluster information
    serialized = re.sub(r"=\d+", "", serialized)
    expectation = re.sub(r"=\d+", "", expectation)
    assert "["+serialized+"]" == expectation



from fontFeatures import Positioning, ValueRecord
from lxml import etree

import unittest


class TestPositioning(unittest.TestCase):
    def roundTrip(self, thing):
        rt = thing.__class__.fromXML(thing.toXML())
        self.assertEqual(rt.asFea(), thing.asFea())

    def test_single(self):
        v = ValueRecord(xAdvance=120)
        s = Positioning(["a"], [v])
        self.assertEqual(s.asFea(), "pos a 120;")
        print(etree.tostring(s.toXML()))
        self.assertEqual(etree.tostring(s.toXML()), "<positioning><glyphs><slot><glyph>a</glyph></slot></glyphs><positions><valuerecord xAdvance=\"120\"/></positions></positioning>".encode("utf-8"))
        self.roundTrip(s)

    def test_single_class(self):
        v = ValueRecord(xPlacement=120)
        s = Positioning([["a", "b"]], [v])
        self.assertEqual(s.asFea(), "pos [a b] <120 0 0 0>;")

    def test_kern(self):
        v = ValueRecord(xAdvance=120)
        s = Positioning(["a", "b"], [v, ValueRecord()])
        self.assertEqual(s.asFea(), "pos a b 120;")



from fontFeatures import FontFeatures, Substitution, Routine
from fontFeatures.shaperLib.Buffer import Buffer
from fontFeatures.shaperLib.Shaper import Shaper
from babelfont import load
import pytest

@pytest.mark.skip("Font too broken to use")
def test_double_application():
    font = load("tests/data/acbe26ce904463c690fb67f70679447059d13ee4.otf" )
    buf = Buffer(font, glyphs=["dvKA","dvVirama","dvKA", "dvVirama", "dvKA"])
    rule = Substitution( [["dvKA"],["dvVirama"]], [["dvK"]] )
    rule.apply_to_buffer(buf)
    assert buf.serialize(position=False) == "dvK|dvK|dvKA"

def test_multiple_applications():
    font = load("tests/data/LibertinusSans-Regular.otf")
    buf = Buffer(font, glyphs=["A", "B", "C"])
    r = Routine()
    r.addRule( Substitution( [["A"]], [["X"]] ) )
    r.addRule( Substitution( [["B"]], [["Y"]] ) )
    r.apply_to_buffer(buf)
    assert buf.serialize(position=False) == "X|Y|C"


def test_namedclass_in_slot():
    font = load("tests/data/LibertinusSans-Regular.otf")
    buf = Buffer(font, glyphs=["A", "B", "C"])
    r = Routine()
    r.addRule( Substitution( [["G", "@AB"]], [["X"]] ) )
    r.apply_to_buffer(buf, namedclasses={"AB": ["A","B"]})
    assert buf.serialize(position=False) == "X|X|C"



from .BaseShaper import BaseShaper
from fontFeatures.shaperLib.Buffer import BufferItem
import unicodedata

LBase = 0x1100
VBase = 0x1161
TBase = 0x11A7
LCount = 19
VCount = 21
TCount = 28
SBase = 0xAC00
NCount = (VCount * TCount)
SCount = (LCount * NCount)

def isCombiningL(u):
    return LBase <= u <= LBase+LCount-1
def isCombiningV(u):
    return VBase <= u <= VBase+VCount-1
def isCombiningT(u):
    return TBase+1 <= u <= TBase+TCount-1
def isCombinedS(u):
    return SBase <= u <= SBase+SCount-1

def isL(u):
    return 0x1100 <= u <= 0x115F or 0xA960 <= u <= 0xA97C
def isV(u):
    return 0x1160 <= u <= 0x11A7 or 0xD7B0 <= u <= 0xD7C6
def isT(u):
    return 0x11A8 <= u <= 0x11FF or 0xD7CB <= u <= 0xD7FB

def isHangulTone(u):
    return 0x302E <= u <= 0x302F

class HangulShaper(BaseShaper):
    def collect_features(self, shaper):
        shaper.add_features("ljmo", "vjmo", "tjmo")

    def override_features(self, shaper):
        shaper.disable_feature("calt")

    def preprocess_text(self):
        def cp(i): return self.buffer.items[i].codepoint
        def has_glyph(u): return self.font.unicode_map.get(u, None)

        # Setup masks here too
        for i in self.buffer.items:
            i.feature_masks["ljmo"] = False
            i.feature_masks["vjmo"] = False
            i.feature_masks["tjmo"] = False
        start, end, processed = 0, 0, 0
        count = len(self.buffer.items)
        for i in range(count):
            u = cp(i)
            if isHangulTone(u):
                if start < end and end == processed:
                    if not self.is_zero_width_char(u):
                        # merge out clusters?

                        # Move end to start
                        l = self.buffer.items
                        l.insert(start, l.pop(end))
                else:
                    if has_glyph(0x25CC):
                        if self.is_zero_width_char(u):
                            l.insert(i+1, BufferItem.new_unicode(0x25CC))
                        else:
                            l.insert(i, BufferItem.new_unicode(0x25CC))
                start = end = i
                continue
            start = i
            if isL(u) and i + 1 < count:
                l = u
                v = cp(i+1)
                if isV(v):
                    # <L,V> or <L,V,T>
                    t, tindex = 0,0
                    if i +2 < count:
                        t = cp(i+2)
                        if isT(t):
                            tindex = t - TBase
                        else:
                            t = 0
                    # Unsafe to break
                    if isCombiningL(l) and isCombiningV(v) and (not t or isCombiningT(t)):
                        s = SBase + (l - LBase) * NCount + (v - VBase) * TCount + tindex;
                        if has_glyph(s):
                            # Combine
                            if t:
                                self.buffer.items[i:i+3] = [BufferItem.new_unicode(s)]
                            else:
                                self.buffer.items[i:i+2] = [BufferItem.new_unicode(s)]
                            end = start + 1
                            continue
                    # Didn't compose
                    self.buffer.items[i].feature_masks["ljmo"] = True
                    self.buffer.items[i+1].feature_masks["vjmo"] = True
                    if t:
                        self.buffer.items[i+2].feature_masks["tjmo"] = True
                        end = start + 3
                    else:
                        end = start + 2
                    continue
            elif isCombinedS(u):
                s = u
                has_glyph_s = has_glyph(s)
                lindex = (s - SBase) // NCount
                nindex = (s - SBase) % NCount
                vindex = nindex // TCount
                tindex = nindex % TCount
                if not tindex and i + 1 < count and isCombiningT(self.buffer.items[i+1].codepoint):
                    # <LV,T>
                    new_tindex = cp(i+1) - TBase
                    new_s = s + new_tindex
                    if has_glyph(new_s):
                        self.buffer.items[i:i+2] = [BufferItem.new_unicode(new_s)]
                        end = start + 1
                        continue
                    else:
                        # unsafe to break
                        pass
                if not has_glyph or (not tindex and i + 1 < count and isT(cp(i+1))):
                    decomposed = [LBase  + lindex, VBase + vindex, TBase + tindex]
                    if has_glyph(decomposed[0]) and has_glyph(decomposed[1]) and (not tindex or has_glyph(decomposed[2])):
                        if not tindex:
                            decomposed.pop()
                        s_len = len(decomposed)
                        self.buffer.items[i:i+s_len] = [BufferItem.new_unicode(u) for u in decomposed]
                        if has_glyph_s and not tindex:
                            i = i + 1
                            s_len = s_len + 1
                        end = start + s_len
                        self.buffer.items[start+1].feature_masks["ljmo"] = True
                        self.buffer.items[start+2].feature_masks["vjmo"] = True
                        if start + 3 < end:
                            self.buffer.items[start+3].feature_masks["tjmo"] = True
                        # merge out clusters
                        continue
                    elif not tindex and i + 1 < count and isT(cp(i+1)):
                        # Mark unsafe to break
                        pass








import logging

__all__ = ["apply_to_buffer"]


def is_forward(rules):
    clearly_forward = False
    clearly_reverse = False
    for r in rules:
        if hasattr(r, "reverse") and r.reverse:
            clearly_reverse = True
        else:
            clearly_forward = True
    if clearly_forward and clearly_reverse:
        logging.getLogger("fontFeatures.shaperLib").warn("There were forward and reverse rules in the same routine!")
        return True
    return not clearly_reverse


def apply_to_buffer(self, buf, stage=None, feature=None, namedclasses={}):
    buf.set_mask(self.flags, self.markFilteringSet, self.markAttachmentSet)
    if feature:
        buf.set_feature_mask(feature)
    # XXX reverse sub routines must go backwards here
    if is_forward(self.rules):
        i = 0
        while i < len(buf): # (which may change!)
            for r in self.rules:
                if stage and r.stage != stage:
                    continue
                if r.flags:
                    buf.set_mask(r.flags, self.markFilteringSet, self.markAttachmentSet)
                if r.would_apply_at_position(buf, i,namedclasses=namedclasses):
                    logging.getLogger("fontFeatures.shaperLib").debug("Applying rule at position %i\n" % (i))
                    delta = r._do_apply(buf, i, namedclasses=namedclasses)
                    buf.update()
                    if delta:
                        i = i + delta
                    break
            i = i + 1
    else:
        i = len(buf)
        while i >= 0:
            for r in self.rules:
                if stage and r.stage != stage:
                    continue
                if r.flags:
                    buf.set_mask(r.flags, self.markFilteringSet, self.markAttachmentSet)
                if r.would_apply_at_position(buf, i,namedclasses=namedclasses):
                    logging.getLogger("fontFeatures.shaperLib").debug("Applying rule at position %i\n" % (i))
                    delta = r._do_apply(buf, i, namedclasses=namedclasses)
                    buf.update()
                    break
            i = i - 1



from fontFeatures import Substitution, FontFeatures, Routine, Chaining, RoutineReference

import pytest


def test_routine_partition():
    f = FontFeatures()

    s1 = Substitution([["A"]],  [["A.grk"]], languages=["grek/*"])
    s2 = Substitution([["A"]],  [["A.esp"]], languages=["latn/ESP "])
    r = Routine(rules=[s1,s2], flags=0x2)

    f.routines.append(r)

    dummy = Routine(rules=[Substitution([["G"]], [["G"]])])
    f.routines.append(dummy)

    c = Chaining([["A"], ["V"]], lookups=[
        [RoutineReference(routine=dummy), RoutineReference(routine=r)],
        [RoutineReference(routine=r), RoutineReference(routine=dummy)],
        ])
    r2 = Routine(rules=[c])
    f.routines.append(r2)

    f.addFeature("locl", [r])

    f.partitionRoutine(r, lambda rule: tuple(rule.languages or []))

    assert len(f.routines) == 4
    assert f.routines[0].flags == f.routines[1].flags
    assert len(f.routines[0].rules) == 1
    assert len(f.routines[1].rules) == 1
    assert f.routines[0].rules[0].replacement[0][0] == "A.grk"
    assert f.routines[1].rules[0].replacement[0][0] == "A.esp"

    assert len(c.lookups[0]) == 3
    assert len(f.features["locl"]) == 2



def test_routine_partition_not_needed():
    f = FontFeatures()

    s1 = Substitution([["A"]],  [["A.grk"]], languages=["grek/*"])
    s2 = Substitution([["A"]],  [["A.esp"]], languages=["grek/*"])
    r = Routine(rules=[s1,s2], flags=0x2)

    f.routines.append(r)
    f.partitionRoutine(r, lambda rule: tuple(rule.languages or []))

    assert len(f.routines) == 1

    r.rules = []
    f.partitionRoutine(r, lambda rule: tuple(rule.languages or []))
    assert len(f.routines) == 1



from fontFeatures import Substitution, Chaining, Routine
import logging


class MergeNonOverlappingRoutines:
    level = 2

    def apply(self, ff):
        logger = logging.getLogger("fontFeatures")
        deadRoutine = Routine(name="dead")

        def haschains(routine):
            for r in routine.rules:
                if isinstance(r, Chaining):
                    return True

        routinelist = ff.routines
        ff.markRoutineUseInChains()
        for lix, l in enumerate(routinelist):
            if haschains(l) or l == deadRoutine:
                continue
            for rix in range(lix + 1, len(routinelist)):
                r = routinelist[rix]
                if haschains(r) or r == deadRoutine:
                    continue
                if self.nonOverlapping(ff, l, r):
                    logger.info(
                        "Merging nonoverlapping routines %s , %s" % (l.name, r.name)
                    )
                    self.merge(l, r)
                    routinelist[rix] = deadRoutine
        ff.routines = list(filter(lambda r: r != deadRoutine, routinelist))

    def whatuses(self, routine, ff):
        # Compute the glyphs which are a context for a routine to be called
        return set.union(set([]), *[c.involved_glyphs for c in list(routine.usedin)])

    def replaceAllReferences(self, usedin, first, second):
        for i in usedin:
            for routinelist in i.lookups:
                if not routinelist:
                    continue
                for ix, routine in enumerate(routinelist):
                    if routine == first:
                        routinelist[ix] = second

    def merge(self, l, r):
        l.rules.extend(r.rules)
        if l.name and r.name:
            l.name = l.name + "_" + r.name
        elif r.name:
            l.name = r.name
        l.comments.extend(r.comments)
        l.usedin = set.union(l.usedin, r.usedin)
        self.replaceAllReferences(l.usedin, r, l)

    def nonOverlapping(self, ff, l, r):
        return (
            not (l.involved_glyphs & r.involved_glyphs)
            and not (self.whatuses(l, ff) & self.whatuses(r, ff))
            and self.compatibleRules(l, r)
        )

    def compatibleRules(self, l, r):
        from fontFeatures.feaLib.Routine import (
            arrange_by_type,
            arrange_by_lookup_type,
            arrange_by_flags,
        )

        testRoutine = Routine()
        testRoutine.rules.extend(l.rules)
        testRoutine.rules.extend(r.rules)
        if arrange_by_type(testRoutine):
            return False
        if arrange_by_lookup_type(testRoutine):
            return False
        if arrange_by_flags(testRoutine):
            return False
        return True


optimizations = [MergeNonOverlappingRoutines]



"""Convert a fontFeatures object to GPOS/GSUB/GDEF tables

This module contains routines for building binary layout tables from a
fontFeatures object. These are implemented as mixins loaded into the
fontFeatures object. e.g. instead of::

    fontFeatures.ttLib.FontFeatures.buildBinaryFeatures(ff, font)

call::

    ff.buildBinaryFeatures(font)

"""
import copy
from fontTools.ttLib.tables import otBase, otTables
from fontTools.ttLib import newTable
from collections import OrderedDict
import itertools
from fontTools.varLib.varStore import OnlineVarStoreBuilder


def buildBinaryFeatures(self, font, axes=[]):
    """Adds GDEF, GSUB and GPOS tables to a font object.

    Args:
        font: a fontTools ``ttFont`` object.
        axes: an optional list of objects conforming to the
          ``fontTools.designspaceLib.AxisDescriptor`` protocol.
    """
    self.resolveAllRoutines()
    reorderRoutines(self)

    # Partition rules based on lookup types
    for r in self.routines:
        self.partitionRoutine(r, lambda rule: rule.lookup_type())

    # XXX first build gdef for mark attachment classes
    if axes:
        self.varstorebuilder = OnlineVarStoreBuilder([ax.tag for ax in axes])
        self.axes = axes
    buildGDEF(self, font)
    buildGPOSGSUB(self, font)
    if axes:
        store = self.varstorebuilder.finish()
        if store.VarData:
            font["GDEF"].table.Version = 0x00010003
            font["GDEF"].table.VarStore = store
            varidx_map = store.optimize()

            font["GDEF"].table.remap_device_varidxes(varidx_map)
            if 'GPOS' in font:
                font['GPOS'].table.remap_device_varidxes(varidx_map)

def reorderRoutines(self):
    """Reorders the routines table to ensure that all routines which are
    referenced in chaining rules appear before the routines which reference them."""
    newroutines = []
    i = 0
    while i < len(self.routines):
        r = self.routines[i]
        if r in newroutines:
            i = i + 1
            continue
        for dep in r.dependencies:
            if dep not in newroutines:
                newroutines.append(dep)
        newroutines.append(r)
        i = i + 1
    self.routines = newroutines

# I am stealing the fontTools.feaLib.builder stuff here
def buildGDEF(self, font):
    """Build a GDEF table and add it to the ``ttFont`` object."""
    gdef = otTables.GDEF()
    gdef.GlyphClassDef = _buildGDEFGlyphClassDef(self)

    # gdef.AttachList = otl.buildAttachList(self.attachPoints_, self.glyphMap)
    # gdef.LigCaretList = otl.buildLigCaretList(
        # self.ligCaretCoords_, self.ligCaretPoints_, self.glyphMap
    # )
    # gdef.MarkAttachClassDef = self.buildGDEFMarkAttachClassDef_()
    # gdef.MarkGlyphSetsDef = self.buildGDEFMarkGlyphSetsDef_()
    # gdef.Version = 0x00010002 if gdef.MarkGlyphSetsDef else 0x00010000
    gdef.Version = 0x00010000
    if any(
        (
            gdef.GlyphClassDef,
            # gdef.AttachList,
            # gdef.LigCaretList,
            # gdef.MarkAttachClassDef,
            # gdef.MarkGlyphSetsDef,
        )
    ):
        font["GDEF"] = newTable("GDEF")
        font["GDEF"].table = gdef


classnames = ["", "base", "ligature", "mark", "component"]

def _buildGDEFGlyphClassDef(self):
    classes = {}
    for g, c in self.glyphclasses.items():
        classes[g] = classnames.index(c)
    if classes:
        result = otTables.GlyphClassDef()
        result.classDefs = classes
        return result
    else:
        return None

def buildGPOSGSUB(self, font):
    """Builds GSUB and GPOS tables and adds them to the ``ttFont`` object."""
    for tag in ["GSUB", "GPOS"]:
        table = makeTable(self, tag, font)
        fontTable = font[tag] = newTable(tag)
        fontTable.table = table


def arrangeByScripts(self):
    """Hoists script/language data from rules into routines, and builds a
    map of which routines apply to each script/language pair.

    Returns an ordered dictionary; the keys are tuples of
    ``(feature_tag, script_tag, language_tag))``, and the values are a list of
    routines."""
    for r in self.routines:
        if any(rule.languages for rule in r.rules):
            self.partitionRoutine(r, lambda rule: tuple(rule.languages or []))

    for r in self.routines:
        if not r.rules:
            continue
        r.languages = r.rules[0].languages # Guaranteed to be the same
        if r.languages:
            for ix,langpair in enumerate(r.languages):
                if langpair[1] == "*":
                    r.languages[ix] = (langpair[0], "dflt")
                if langpair[0] == "*":
                    r.languages[ix] = ("DFLT", langpair[1])
        else:
            r.languages = [("DFLT", "dflt")]

    self.hoist_languages()
    script_lang_pairs = []
    for script in self.scripts_and_languages.keys():
        if not (script, "dflt") in script_lang_pairs:
            script_lang_pairs.append((script, "dflt"))
        for lang in self.scripts_and_languages[script]:
            script_lang_pairs.append((script, lang))

    the_big_map = OrderedDict()
    def put_in_map(tag, script, lang, routine):
        key = (tag, script, lang)
        if key not in the_big_map:
            the_big_map[key] = []
        if routine not in the_big_map[key]:
            the_big_map[key].append(routine)

    def put_in_map_with_default(tag, script, lang, routine):
        put_in_map(tag, script, lang, routine)
        for script2, lang2 in script_lang_pairs:
            if script == "DFLT" and lang == "dflt":
                put_in_map(tag, script2, lang2, routine)
            elif script == script2 and lang == "dflt":
                put_in_map(tag, script2, lang2, routine)

    for tag, routinereferences in self.features.items():
        for r in routinereferences:
            for script, lang in r.routine.languages:
                put_in_map_with_default(tag, script, lang, r)
    return the_big_map

def separate_by_stage(the_big_map, stage):
    """Splits up the routine map into GSUB/GPOS routines."""
    stage_map = OrderedDict()
    for k,v in the_big_map.items():
        v = [r for r in v if r.stage == stage]
        if v:
            stage_map[k] = v
    return stage_map

def makeTable(self, tag, font):
    """Compiles a binary GSUB/GPOS table."""
    table = getattr(otTables, tag, None)()
    table.Version = 0x00010000
    table.ScriptList = otTables.ScriptList()
    table.ScriptList.ScriptRecord = []
    table.FeatureList = otTables.FeatureList()
    table.FeatureList.FeatureRecord = []
    table.LookupList = otTables.LookupList()

    stage_map = separate_by_stage(arrangeByScripts(self), tag[1:].lower())
    stage_routines = [x for x in self.routines if x.stage == tag[1:].lower() ]
    buildersset = [ x.toOTLookup(font, self) for x in stage_routines ]
    lookups = []
    builderlist = []
    for builders in buildersset:
        for builder in builders:
            builder.lookup_index = len(lookups)
            builderlist.append(builder)
            lookups.append(builder.build())

    table.LookupList.Lookup = lookups

    # Build a table for mapping (tag, lookup_indices) to feature_index.
    # For example, ('liga', (2,3,7)) --> 23.
    feature_indices = {}
    required_feature_indices = {}  # ('latn', 'DEU') --> 23
    scripts = {}  # 'latn' --> {'DEU': [23, 24]} for feature #23,24
    # Sort the feature table by feature tag:
    # https://github.com/fonttools/fonttools/issues/568
    sortFeatureTag = lambda f: (f[0][2], f[0][1], f[0][0], f[1])
    for key, lookups in sorted(stage_map.items(), key=sortFeatureTag):
        feature_tag, script, lang = key
        # l.lookup_index will be None when a lookup is not needed
        # for the table under construction. For example, substitution
        # rules will have no lookup_index while building GPOS tables.
        lookup_indices = tuple([builderlist.index(x.routine.__builder) for x in lookups ])

        size_feature = tag == "GPOS" and feature_tag == "size"
        if len(lookup_indices) == 0 and not size_feature:
            continue

        feature_key = (feature_tag, lookup_indices)
        feature_index = feature_indices.get(feature_key)
        if feature_index is None:
            feature_index = len(table.FeatureList.FeatureRecord)
            frec = otTables.FeatureRecord()
            frec.FeatureTag = feature_tag
            frec.Feature = otTables.Feature()
            # frec.Feature.FeatureParams = self.buildFeatureParams(feature_tag)
            frec.Feature.LookupListIndex = list(lookup_indices)
            frec.Feature.LookupCount = len(lookup_indices)
            table.FeatureList.FeatureRecord.append(frec)
            feature_indices[feature_key] = feature_index
        scripts.setdefault(script, {}).setdefault(lang, []).append(feature_index)

    # Build ScriptList.
    for script, lang_features in sorted(scripts.items()):
        srec = otTables.ScriptRecord()
        srec.ScriptTag = script
        srec.Script = otTables.Script()
        srec.Script.DefaultLangSys = None
        srec.Script.LangSysRecord = []
        for lang, feature_indices in sorted(lang_features.items()):
            langrec = otTables.LangSysRecord()
            langrec.LangSys = otTables.LangSys()
            langrec.LangSys.LookupOrder = None
            langrec.LangSys.ReqFeatureIndex = 0xFFFF
            langrec.LangSys.FeatureIndex = [i for i in feature_indices ]
            langrec.LangSys.FeatureCount = len(langrec.LangSys.FeatureIndex)
            if lang == "dflt":
                srec.Script.DefaultLangSys = langrec.LangSys
            else:
                langrec.LangSysTag = lang
                srec.Script.LangSysRecord.append(langrec)
        srec.Script.LangSysCount = len(srec.Script.LangSysRecord)
        table.ScriptList.ScriptRecord.append(srec)

    table.ScriptList.ScriptCount = len(table.ScriptList.ScriptRecord)
    table.FeatureList.FeatureCount = len(table.FeatureList.FeatureRecord)
    table.LookupList.LookupCount = len(table.LookupList.Lookup)
    return table



from collections import OrderedDict
from enum import IntEnum
import re
from youseedee import ucd_data


script_config = {
  "Invalid":    dict(old_spec= False, virama =      0, base_pos = "last", reph_pos = "before_post", reph_mode = "implicit", blwf_mode =  "pre_and_post"),
  "Devanagari": dict(old_spec= True,  virama = 0x094D, base_pos = "last", reph_pos = "before_post", reph_mode = "implicit", blwf_mode =  "pre_and_post"),
  "Bengali":    dict(old_spec= True,  virama = 0x09CD, base_pos = "last", reph_pos = "after_sub", reph_mode =   "implicit", blwf_mode =  "pre_and_post"),
  "Gurmukhi":   dict(old_spec= True,  virama = 0x0A4D, base_pos = "last", reph_pos = "before_sub", reph_mode =  "implicit", blwf_mode =  "pre_and_post"),
  "Gujarati":   dict(old_spec= True,  virama = 0x0ACD, base_pos = "last", reph_pos = "before_post", reph_mode = "implicit", blwf_mode =  "pre_and_post"),
  "Oriya":      dict(old_spec= True,  virama = 0x0B4D, base_pos = "last", reph_pos = "after_main", reph_mode =  "implicit", blwf_mode =  "pre_and_post"),
  "Tamil":      dict(old_spec= True,  virama = 0x0BCD, base_pos = "last", reph_pos = "after_post", reph_mode =  "implicit", blwf_mode =  "pre_and_post"),
  "Telugu":     dict(old_spec= True,  virama = 0x0C4D, base_pos = "last", reph_pos = "after_post", reph_mode =  "explicit", blwf_mode =  "post_only"),
  "Kannada":    dict(old_spec= True,  virama = 0x0CCD, base_pos = "last", reph_pos = "after_post", reph_mode =  "implicit", blwf_mode =  "post_only"),
  "Malayalam":  dict(old_spec= True,  virama = 0x0D4D, base_pos = "last", reph_pos = "after_main", reph_mode =  "log_repha", blwf_mode = "pre_and_post"),
  "Sinhala":    dict(old_spec= False, virama = 0x0DCA, base_pos = "last_sinhala", reph_pos = "after_post", reph_mode =  "explicit", blwf_mode = "pre_and_post")
}

# Go from UCD categories to the ones we care about

syllabic_category_map = {
  "Other": "X",
  "Avagraha": "Symbol",
  "Bindu": "SM",
  "Brahmi_Joining_Number": "PLACEHOLDER", # Don't care.
  "Cantillation_Mark": "A",
  "Consonant": "C",
  "Consonant_Dead": "C",
  "Consonant_Final": "CM",
  "Consonant_Head_Letter": "C",
  "Consonant_Killer": "M", # U+17CD only.
  "Consonant_Medial": "CM",
  "Consonant_Placeholder": "PLACEHOLDER",
  "Consonant_Preceding_Repha": "Repha",
  "Consonant_Prefixed": "X", # Don't care.
  "Consonant_Subjoined": "CM",
  "Consonant_Succeeding_Repha": "CM",
  "Consonant_With_Stacker": "CS",
  "Gemination_Mark": "SM", # https://github.com/harfbuzz/harfbuzz/issues/552
  "Invisible_Stacker": "Coeng",
  "Joiner": "ZWJ",
  "Modifying_Letter": "X",
  "Non_Joiner": "ZWNJ",
  "Nukta": "N",
  "Number": "PLACEHOLDER",
  "Number_Joiner": "PLACEHOLDER", # Don't care.
  "Pure_Killer": "M", # Is like a vowel matra.
  "Register_Shifter": "RS",
  "Syllable_Modifier": "SM",
  "Tone_Letter": "X",
  "Tone_Mark": "N",
  "Virama": "H",
  "Visarga": "SM",
  "Vowel": "V",
  "Vowel_Dependent": "M",
  "Vowel_Independent": "V",
  # The following come from reassignment
  "Symbol": "Symbol",
  "Ra": "Ra",
  "Dotted_Circle": "DOTTEDCIRCLE",
}

## Set up the state machine for syllable matching and syllabic
## categories using composed regular expressions


states = OrderedDict(
    c = "C|Ra",
    n = "(ZWNJ?RS)?(N N?)?",
    z = "ZWJ|ZWNJ",
    reph = "Ra H|Repha",
    cn = "c ZWJ?n?",
    forced_rakar = "ZWJ H ZWJ Ra",
    symbol = "Symbol N?",
    matra_group = "z*M N?(H|forced_rakar)?",
    syllable_tail = "(z?SM SM?ZWNJ?)?A*",
    halant_group = "z?H(ZWJ N?)?",
    final_halant_group = "halant_group|H ZWNJ",
    medial_group = "CM?",
    halant_or_matra_group = "final_halant_group|matra_group*",
    complex_syllable_tail = "(halant_group cn)*medial_group halant_or_matra_group syllable_tail",
    consonant_syllable = "((Repha|CS)?cn complex_syllable_tail)",
    vowel_syllable = "(reph?V n?(complex_syllable_tail|ZWJ))",
    standalone_cluster = "((Repha|CS)?PLACEHOLDER|reph?DOTTEDCIRCLE)n? complex_syllable_tail",
    symbol_cluster = "symbol syllable_tail",
    broken_cluster = "reph? n? complex_syllable_tail",
)

def make_syllable_machine(states, additional_categories=[]):
    categories = set(syllabic_category_map.values()) | set(additional_categories)
    syllable_machine = {}
    tail = r'\([^\)]+\)=\d+'
    for value in categories:
        syllable_machine[value] = f'<{value}>' + tail
    syllable_machine["other"] = r'<[^\>]+>' + tail

    for category,definition in states.items():
        # Recursively replace any known states with their definitions:
        states = "\\b("+"|".join(sorted(syllable_machine.keys(),key=len, reverse=True))+")\\b"
        definition = re.sub(states, lambda match: "("+syllable_machine[match[1]]+")", definition)
        definition = definition.replace(" ", "")
        syllable_machine[category] = definition
    return syllable_machine

syllable_machine_indic = make_syllable_machine(states)

class IndicPosition(IntEnum):
  START = 0

  RA_TO_BECOME_REPH = 1
  PRE_M = 2
  PRE_C = 3

  BASE_C = 4
  AFTER_MAIN = 5

  ABOVE_C = 6

  BEFORE_SUB = 7
  BELOW_C = 8
  AFTER_SUB = 9

  BEFORE_POST = 10
  POST_C = 11
  AFTER_POST = 12

  FINAL_C = 13
  SMVD = 14

  END = 15

def is_ra(u):
    return u in [0x0930, 0x09B0, 0x09F0, 0x0A30, 0x0AB0, 0x0B30, 0x0BB0, 0x0C30, 0x0CB0, 0x0D30, 0x0DBB, 0x179A]

def IndicPositionalCategory2IndicPosition(ipc):
    if ipc == "Left":   return IndicPosition.PRE_C
    if ipc == "Top":    return IndicPosition.ABOVE_C
    if ipc == "Bottom": return IndicPosition.BELOW_C

    if ipc == "Right":  return IndicPosition.POST_C
    if ipc == "Bottom_And_Right": return IndicPosition.POST_C
    if ipc == "Left_And_Right": return IndicPosition.POST_C
    if ipc == "Top_And_Bottom_And_Left": return IndicPosition.POST_C

    if ipc == "Top_And_Bottom": return IndicPosition.BELOW_C
    if ipc == "Top_And_Bottom_And_Left": return IndicPosition.BELOW_C
    if ipc == "Top_And_Bottom_And_Right": return IndicPosition.POST_C

    if ipc == "Top_And_Left": return IndicPosition.ABOVE_C
    if ipc == "Top_And_Left_And_Right": return IndicPosition.POST_C
    if ipc == "Top_And_Right": return IndicPosition.POST_C

    if ipc == "Overstruck": return IndicPosition.AFTER_MAIN
    if ipc == "Visual_Order_Left": return IndicPosition.PRE_M

    return IndicPosition.END

matra_pos_left = { "Default": IndicPosition.PRE_M }
matra_pos_right = {
# Telu and Knda need special cases
    "Devanagari": IndicPosition.AFTER_SUB,
    "Bengali": IndicPosition.AFTER_POST,
    "Gurmukhi": IndicPosition.AFTER_POST,
    "Gujarati": IndicPosition.AFTER_POST,
    "Oriya": IndicPosition.AFTER_POST,
    "Tamil": IndicPosition.AFTER_POST,
    "Malayalam": IndicPosition.AFTER_POST,
    "Default": IndicPosition.AFTER_SUB
}
matra_pos_top = {
    "Devanagari": IndicPosition.AFTER_SUB,
    "Gurmukhi": IndicPosition.AFTER_POST,
    "Gujarati": IndicPosition.AFTER_SUB,
    "Oriya": IndicPosition.AFTER_MAIN,
    "Tamil": IndicPosition.AFTER_SUB,
    "Telugu": IndicPosition.BEFORE_SUB,
    "Kannada": IndicPosition.BEFORE_SUB,
    "Sinhala": IndicPosition.AFTER_SUB,
    "Default": IndicPosition.AFTER_SUB
}
matra_pos_bottom = {
    "Devanagari": IndicPosition.AFTER_SUB,
    "Bengali": IndicPosition. AFTER_SUB,
    "Gurmukhi": IndicPosition.AFTER_POST,
    "Gujarati": IndicPosition.AFTER_POST,
    "Oriya": IndicPosition.AFTER_SUB,
    "Tamil": IndicPosition.AFTER_POST,
    "Telugu": IndicPosition.BEFORE_SUB,
    "Malayalam": IndicPosition.AFTER_POST,
    "Sinhala": IndicPosition.AFTER_SUB,
    "Default": IndicPosition.AFTER_SUB
}

def set_matra_position(item):
    script = ucd_data(item.codepoint)["Script"]
    u = item.codepoint
    if item.syllabic_position == IndicPosition.PRE_C:
        selector = matra_pos_left
    elif item.syllabic_position == IndicPosition.POST_C:
        selector = matra_pos_right
        if script == "Telugu":
            if u <= 0x0C42:
                item.syllabic_position = IndicPosition.BEFORE_SUB
            else:
                item.syllabic_position = IndicPosition.AFTER_SUB
            return
        if script == "Kannada":
            if u < 0x0CC3 or u > 0xCD6:
                item.syllabic_position = IndicPosition.BEFORE_SUB
            else:
                item.syllabic_position = IndicPosition.AFTER_SUB
            return
    elif item.syllabic_position == IndicPosition.ABOVE_C:
        selector = matra_pos_top
    elif item.syllabic_position == IndicPosition.BELOW_C:
        selector = matra_pos_bottom
    else:
        return
    item.syllabic_position = selector.get(script, selector["Default"])

def reassign_category_and_position_indic(item):
    cp = item.codepoint
    if 0x0953 <= cp <= 0x0954:
        item.syllabic_category = "SM"
    elif 0x0A72 <= cp <= 0x0A73 or 0x1CF5 <= cp <= 0x1CF6:
        item.syllabic_category = "C"
    elif 0x1CE2 <= cp <= 0x1CE8 or cp == 0x1CED:
        item.syllabic_category = "A"
    elif 0xA8F2 <= cp <= 0xA8F7 or 0x1CE9 <= cp <= 0x1CEC or 0x1CEE <= cp <= 0x1CF1:
        item.syllabic_category = "Symbol"
    elif cp == 0x0A51: # https://github.com/harfbuzz/harfbuzz/issues/524
        item.syllabic_category = "M"
        item.syllabic_position = IndicPosition.BELOW_C
    elif cp == 0x11301 or cp == 0x11303:
        item.syllabic_category = "SM"
    elif cp == 0x1133B or cp == 0x1133C:
        item.syllabic_category = "N"
    elif cp == 0x0AFB:
        item.syllabic_category = "N"
    elif cp == 0x0980 or cp == 0x09FC or cp == 0x0C80 or cp == 0x2010 or cp == 0x2011:
        item.syllabic_category = "PLACEHOLDER"
    elif cp == 0x25CC:
        item.syllabic_category = "DOTTEDCIRCLE"

    isc = item.syllabic_category

    is_medial = isc == "CM"
    is_consonant = isc in ["C", "CS", "Ra", "V", "PLACEHOLDER", "DOTTEDCIRCLE"] or is_medial

    if is_consonant:
        item.syllabic_position = IndicPosition.BASE_C
        if is_ra(cp):
            item.syllabic_category = "Ra"
    elif isc == "M":
        set_matra_position(item)
    elif isc in ["SM", "A", "Symbol"]:
        item.syllabic_position = IndicPosition.SMVD
    if cp == 0x0B01:
        item.syllabic_position = IndicPosition.BEFORE_SUB



"""Routines for converting Rule objects to and from XML."""
from lxml import etree
from fontFeatures.xmlLib.utils import put_languages, put_address

@classmethod
def fromXML(klass, el):
    """Creates a Rule from a lxml Element object."""
    import fontFeatures

    subklass = getattr(fontFeatures, el.tag.title())
    assert subklass
    return subklass.fromXML(el)


def toXML(self):
    """Serializes a Rule to a lxml Element object."""
    root = etree.Element(self.__class__.__name__.lower())
    put_address(self, root)
    put_languages(self, root)
    if hasattr(self, "reverse") and self.reverse:
        root.attrib["reverse"] = "true"
    if hasattr(self, "flags") and self.flags:
        root.attrib["flags"] = str(self.flags)
    if hasattr(self, "precontext"):
        self._makeglyphslots(root, "precontext", self.precontext)
    if hasattr(self, "postcontext"):
        self._makeglyphslots(root, "postcontext", self.postcontext)
    if hasattr(self, "lookups") and self.lookups:
        wrapper = etree.SubElement(root, "lookups")
        for slot in self.lookups:
            xmlslot = etree.SubElement(wrapper, "slot")
            if slot:
                for lu in slot:
                    xmlslot.append(lu.toXML())
            else:
                etree.SubElement(xmlslot, "lookup")

    return self._toXML(root)


def _makeglyphslots(self, root, tag, list_of_lists):
    if not list_of_lists:
        return
    wrapper = etree.SubElement(root, tag)
    for slot in list_of_lists:
        xmlslot = etree.SubElement(wrapper, "slot")
        for g in slot:
            etree.SubElement(xmlslot, "glyph").text = g


def _slotArray(self, el):
    if el is None:
        return None
    return [[g.text for g in slot.findall("glyph")] for slot in list(el)]



# Code for converting a Routine object into feaLib statements
import fontTools.feaLib.ast as feaast
from fontFeatures.ttLib.Substitution import lookup_type as sub_lookup_type
from fontFeatures.ttLib.Positioning import lookup_type as pos_lookup_type
from fontFeatures.feaLib import bad_statement_to_comment
import copy


def lookup_type(rule):
    from fontFeatures import Substitution, Positioning, Attachment, Chaining

    if isinstance(rule, Substitution):
        return sub_lookup_type(rule)
    if isinstance(rule, Positioning):
        return pos_lookup_type(rule)
    if isinstance(rule, Attachment):
        return rule.is_cursive
    if isinstance(rule, Chaining):
        return 1
    raise ValueError


counter = 0


def gensym(prefix):
    global counter
    counter = counter + 1
    return prefix + str(counter)

def feaPreamble(self, ff):
    preamble = []
    for r in self.rules:
        preamble.extend(r.feaPreamble(ff))
    if self.flags & 0xFF00:
        assert(self.markAttachmentSet is not None)
        self.markAttachmentSetAsClass = ff.getNamedClassFor(self.markAttachmentSet, gensym("markAttachmentSet"))
    if self.flags & 0x10:
        assert(self.markFilteringSet is not None)
        self.markFilteringSetAsClass = ff.getNamedClassFor(self.markFilteringSet, gensym("markFilteringSet"))
    return preamble


def asFeaAST(self):
    if self.name:
        f = feaast.LookupBlock(name=self.name)
    else:
        f = feaast.Block()

    if hasattr(self, "flags"):
        flags = feaast.LookupFlagStatement(self.flags)
        if self.flags & 0x10 and hasattr(self, "markFilteringSetAsClass"): # XXX
            # We only need the name, not the contents
            mfs = feaast.GlyphClassDefinition(self.markFilteringSetAsClass,
                feaast.GlyphClass([])
            )
            flags.markFilteringSet = feaast.GlyphClassName(mfs)
        if self.flags & 0xFF00 and hasattr(self, "markAttachmentSetAsClass"): # XXX
            mfs = feaast.GlyphClassDefinition(self.markAttachmentSetAsClass,
                feaast.GlyphClass([])
            )
            flags.markAttachment=feaast.GlyphClassName(mfs)

        f.statements.append(flags)

    for x in self.comments:
        f.statements.append(feaast.Comment(x))

    f.statements.append(feaast.Comment(";"))
    lastaddress = self.address
    if lastaddress:
        f.statements.append(feaast.Comment("# Original source: %s " % (" ".join([str(x) for x in lastaddress]))))
    for x in self.rules:
        if x.address and x.address != lastaddress:
            f.statements.append(feaast.Comment("# Original source: %s " % x.address))
            lastaddress = x.address
        if hasattr(x, "note"):
            f.statements.append(feaast.Comment("\n".join([ f"# {n}" for n in x.note.split("\n")])))
        f.statements.append(bad_statement_to_comment(x.asFeaAST()))
    return f


def asFea(self):
    return self.asFeaAST().asFea()



"""
pathUtils
=========

This is a set of simple routines to help with extracting information from
the outlines of glyphs.
"""
from beziers.path import BezierPath
from beziers.line import Line
from beziers.point import Point


def get_bezier_paths(font, glyphname):
    """Retrieve beziers from a glyph

    Args:
        font: A Babelfont font

    Returns:
        An array of ``beziers.path.BezierPath`` objects representing the
        outlines of the glyph.
    """
    layer = font.default_master.get_glyph_layer(glyphname)
    layer.decompose()
    return BezierPath.fromDrawable(layer, glyphSet = { k:font.default_master.get_glyph_layer(k)
        for k in font.exportedGlyphs()
    })


def find_largest_path(font, glyphname):
    """Find largest path by area

    Args:
        font: A fontTools ``TTFont`` object
        glyphname: The name of a single glyph

    Returns:
        A ``beziers.path.BezierPath`` object representing the largest path
        in the glyph.
    """
    paths = get_bezier_paths(font, glyphname)
    return max(paths, key=lambda p: p.area)


def thickness_at_x(path, x):
    """Find the path thickness at a given X coordinate

    This measure the thickness of the lowest horizontal stem at the given
    coordinate. If there is no stem at this X coordinate, ``None`` is
    returned.

    Args:
        path: A ``beziers.path.BezierPath`` object
        x: X coordinate to search

    Returns:
        The thickness of the path at this point, in font units.
    """
    bounds = path.bounds()
    bounds.addMargin(10)
    ray = Line(Point(x - 0.1, bounds.bottom), Point(x + 0.1, bounds.top))
    intersections = []
    for seg in path.asSegments():
        intersections.extend(seg.intersections(ray))
    if len(intersections) < 2:
        return None
    intersections = list(sorted(intersections, key=lambda i: i.point.y))
    i1, i2 = intersections[0:2]
    inorm1 = i1.seg1.normalAtTime(i1.t1)
    ray1 = Line(i1.point + (inorm1 * 1000), i1.point + (inorm1 * -1000))
    iii = i2.seg1.intersections(ray1)
    if iii:
        ll1 = i1.point.distanceFrom(iii[0].point)
    else:
        # Simple, vertical version
        return abs(i1.point.y - i2.point.y)

    inorm2 = i2.seg1.normalAtTime(i2.t1)
    ray2 = Line(i2.point + (inorm2 * 1000), i2.point + (inorm2 * -1000))
    iii = i1.seg1.intersections(ray2)
    if iii:
        ll2 = i2.point.distanceFrom(iii[0].point)
        return (ll1 + ll2) * 0.5
    else:
        return ll1

    # midpoint = (i1.point + i2.point) / 2
    # # Find closest path to midpoint
    # # Find the tangent at that time
    # inorm2 = i2.seg1.normalAtTime(i2.t1)


# from fontTools.ttLib import TTFont

# font = TTFont("fonts/Amiri-Regular.ttf")
# p = find_largest_path(font, "aHaa.medi")
# print(thickness_at_x(p, 128))



"""GSUBUnparser: Convert binary GSUB lookups to fontFeatures objects."""
import fontTools
from collections import OrderedDict
from .GTableUnparser import GTableUnparser
from itertools import groupby
import fontFeatures

# These are silly little functions which help to document the intent
def glyph(x):
    """Wraps a glyph name in an array, to document the fact that it occupies a slot."""
    assert isinstance(x, str)
    return [x]

def singleglyph(x):
    """Wraps a glyph name in two arrays, to document the fact that it is the only occupant of a slot."""
    return [glyph(x)]


class GSUBUnparser(GTableUnparser):
    """Unparse a GSUB table into a fontFeatures object. See :py:class:`fontFeatures.ttLib.GTableUnparser`."""
    _table = "GSUB"
    lookupTypes = {
        1: "SingleSubstitution",
        2: "MultipleSubstitution",
        3: "AlternateSubstitution",
        4: "LigatureSubstitution",
        5: "Contextual",
        6: "ChainedContextual",
        7: "Extension",
        8: "ReverseContextualSubstitution",
    }

    _attrs = {
        "lookup": "SubstLookupRecord",
        "format1_ruleset": "SubRuleSet",
        "format1_rule": "SubRule",
        "format2_classset": "SubClassSet",
        "format2_rule": "SubClassRule",
        "chain_format1_ruleset": "ChainSubRuleSet",
        "chain_format1_rule": "ChainSubRule",
        "chain_format2_classset": "ChainSubClassSet",
        "chain_format2_rule": "ChainSubClassRule",
    }

    def isChaining(self, lookupType):
        """Returns true if the given lookup type is a chaining lookup."""
        return lookupType >= 5

    def unparseReverseContextualSubstitution(self, lookup):
        """Turn a GPOS8 (reverse contextual substitution) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(
            name=self.getname("ReverseContextualSubstitution" + self.gensym())
        )
        self._fix_flags(b, lookup)
        for sub in lookup.SubTable:
            prefix  = []
            outputs = []
            suffix  = []
            if hasattr(sub, "BacktrackCoverage"):
                for coverage in reversed(sub.BacktrackCoverage):
                    prefix.append(coverage.glyphs)
            if hasattr(sub, "LookAheadCoverage"):
                for i, coverage in enumerate(sub.LookAheadCoverage):
                    suffix.append(coverage.glyphs)
            outputs = [ sub.Substitute ]
            inputs =  [ sub.Coverage.glyphs ]
            b.addRule(
                fontFeatures.Substitution(
                    inputs,
                    outputs,
                    prefix,
                    suffix,
                    flags=lookup.LookupFlag,
                    reverse=True
                )
            )
        return b,[]

    def unparseLigatureSubstitution(self, lookup):
        """Turn a GPOS4 (ligature substitution) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(
            name=self.getname("LigatureSubstitution" + self.gensym())
        )
        self._fix_flags(b, lookup)
        for sub in lookup.SubTable:
            for first, ligatures in sub.ligatures.items():
                for lig in ligatures:
                    substarray = [glyph(first)]
                    for x in lig.Component:
                        substarray.append(glyph(x))
                    b.addRule(
                        fontFeatures.Substitution(
                            substarray,
                            singleglyph(lig.LigGlyph),
                            address=self.currentLookup,
                            flags=lookup.LookupFlag,
                        )
                    )
        return b, []

    def unparseMultipleSubstitution(self, lookup):
        """Turn a GPOS2 (multiple substitution) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(
            name=self.getname("MultipleSubstitution" + self.gensym())
        )
        self._fix_flags(b, lookup)

        for sub in lookup.SubTable:
            for in_glyph, out_glyphs in sub.mapping.items():
                b.addRule(
                    fontFeatures.Substitution(
                        singleglyph(in_glyph),
                        [glyph(x) for x in out_glyphs],
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                )
        return b, []

    def unparseAlternateSubstitution(self, lookup):
        """Turn a GPOS3 (alternate substitution) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(
            name=self.getname("AlternateSubstitution" + self.gensym())
        )
        self._fix_flags(b, lookup)
        for sub in lookup.SubTable:
            for in_glyph, out_glyphs in sub.alternates.items():
                b.addRule(
                    fontFeatures.Substitution(
                        singleglyph(in_glyph),
                        [out_glyphs],
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                        force_alt=True,
                    )
                )
        return b, []

    def unparseSingleSubstitution(self, lookup):
        """Turn a GPOS1 (single substitution) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(
            name=self.getname("SingleSubstitution" + self.gensym())
        )
        self._fix_flags(b, lookup)
        for sub in lookup.SubTable:
            for k, v in sub.mapping.items():
                b.addRule(
                    fontFeatures.Substitution(
                        [[k]],
                        [[v]],
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                )
        return b, []



"""Routines for converting RoutineReference objects to and from XML."""
from lxml import etree


def toXML(self):
    """Serializes a RoutineReference to a lxml Element object."""
    root = etree.Element("routinereference")
    if self.name:
        root.attrib["name"] = self.name
    return root


@classmethod
def fromXML(klass, el):
    """Creates a RoutineReference from a lxml Element object."""
    rule = klass(name=el.get("name"))
    id = el.get("id")
    return rule



from __future__ import print_function, division, absolute_import
from fontTools.feaLib.builder import addOpenTypeFeatures
from fontTools.ttLib import TTFont
import sys

from argparse import ArgumentParser, FileType

parser = ArgumentParser()
parser.add_argument("input",
                    help="font file to process", metavar="OTF")
parser.add_argument('feature', type=FileType('r'), default='-',nargs='?',
                    help="feature file(s) to add", metavar="FEA")
parser.add_argument("-o", dest="output",
                    help="path to output font", metavar="FILE")
args = parser.parse_args()
output = args.output
if output is None:
  output = "fea-"+args.input

font = TTFont(args.input)
print("Adding features")
addOpenTypeFeatures(font, args.feature)
print("Saving on "+output)
font.save(output)



"""ttLib.Substitution: Converting Substitution rules to TrueType."""

def lookup_type(self, forFea=False):
    """Mixin to determine the GSUB lookup type of a fontFeatures.Substitution object

    Returns: integer GSUB lookup type."""
    if self.reverse:
        return 8
    if not self.replacement:
        return 2 # Deletion but treat as mult sub
    if (self.has_context and not forFea) or len(self.lookups) > 0 and any([x is not None for x in self.lookups]):
        return 6  # Chaining
    # if self.input == self.replacement: # It's an ignore
    # return 6
    if len(self.input) == 1 and len(self.replacement) == 1:
        if self.force_alt or (len(self.input[0]) == 1 and len(self.replacement[0]) > 1):
            return 3  # Alternate
        else:
            return 1  # Single
    if len(self.input) > 1 and len(self.replacement) == 1:
        return 4  # Ligature

    if len(self.input) > 1 and len(self.replacement) > 1:
        return 9  # Not directly expressible!

    if len(self.replacement) > 1:
        return 2
